	; FILE EXPORT PROGRAM for CMM2 CP/M simulator
	; This program will copy files within the CP/M systen and also export to the CMM2 SD card.
	; It is based on the program in the Digital Research document 'CP/M 2.0 Interface Guide' (1979) page 30 'A Sample File-to-File Copy Program'.
	; Original modified by GerryL (TBS) in Aug 2022 to also Export files from CP/M A: drive to the CMM2 SD disk, ensure some file checking prior to copy
	; and change the command file position so Destination is on left (first file entered) and Source is on right (second file entered) 
	; as per normal usage.
	;
	; *** CP/M internal File copy ***
	; It is preferred that PIP.COM is used for copying CP/M files
	; command is EXPORT DEST_FILE SOURCE_FILE, e.g COPY Test2.txt Test1.txt (creates Test2.txt and copies contents of Test1.txt to Test2.txt).
	;   If the destination file exists the user will be asked if an overwite is to occur (Y) or if not (N) program will end with no copy action.
	;  *** CP/M file exporting to SD disk in CMM2 ***
	; command is EXPORT SDA: SOURCE_FILE, e.g. EXPORT SDA: Test1.txt will create or overwrite the file CPM_EXPORT.txt on the CMM2 SD disk.  
	; Note the MMBasic program of the Export process treates text file and binary files differently in how they are stored. 
	; Also text file transfer is ended with EOF (CTRL Z/01AH) while a binary file transfer ends when CP/M indicates its the end of file.
	; If the CP/M file type is .txt,.bas, .hex, .asm or .prn (i.e. CP/M text files) then it will save to CPM_EXPORT.txt else it will be treated as a binary file
	; and be saved to CPM_EXPORT.bin. Note that SDA: has a colon at the end which which tells CP/M its a device and not a file.
	;
	; History
	; Version 1. Completed 01 Sep 22
	
	
	OPTIONS \8080 ; using 8080/85 mnemonics
	BOOT equ 0000h ; system reboot
	BDOS equ 0005h ; BDOS entry point
	FCB1 equ 005ch ; first file name dest
	FCBDEST equ FCB1 ; dest fcb 
	FCB2 equ 006ch ; second file name, source
	TPA equ 0100h ; beginning of tpa
	TRUE equ 0ffh
	FALSE equ NOT TRUE
	CPM_OPEN_EXPORT equ 080H ; open export file
	CPM_TRANSFER_EXPORT EQU 081H ;transfers 128 byes
	CPM_CLOSE_EXPORT EQU 0x82 ;closes CPM_EXPORT 
	CPM_CONIN EQU 023H

	
 ;	;BDOS function calls
	printf equ 9 ; print buffer func#
	openf equ 15 ; open file func#
	closef equ 16 ; close file func#
	deletef equ 19 ; delete file func#
	readf equ 20 ; sequential read func#
	writef equ 21 ; sequential write
	makef equ 22; make file func#
	readconbuff equ 10 ; read buffer in
 ;
 ; ================= START =================
	org TPA ; beginning of tpa
	lxi sp,stack ; set local stack
	call IsExport ; ExportFlag=ff if its output to SDA:
	lda ExportFlag
	cpi TRUE
	jz Lab1 ; skip checking dest file name
	; check file names are valid with no ambigous stuff like ? and * or totally missing
	lxi h,FCBDEST
	call ValidName ; check not ambiguous
	Lab1:
	lxi h,FCB2
	call ValidName
	call ValidSrcDest ; ensure source and dest not the same
	; move second file name from FCB2 to FCBSrc
	mvi c, 16 ; half an fcb
	lxi d,FCB2 ; source of move
	lxi h,FCBSrc ; destination fcb
	mfcb: ldax d ; source fcb
	inx d ; ready next
	mov m,a ; dest fcb
	inx h ; ready next
	dcr c ; count 16...0
	jnz mfcb ; loop 16 times
	; name has been removed, zero cr
	xra a ; a = 00h
	sta dSrccbcr
	; source and destination fcb's ready
 	lxi d,FCBSrc ; source
	call open ; error if 255
	lxi d,nofile ; ready message
	inr a ; 255 becomes 0
	cz ProgExit ; exit if no file
	lda ExportFlag
	cpi TRUE
	jz DoExport ; its a export
	; source file open, prep destination
	; first check if overwritting 
	lxi d,FCBDEST
	call close ;return ff if not found
	cpi 0FFH
	jz DoDelete
	lxi d, FileExistMSG
	mvi c,printf
	call BDOS ; write message
	lxi d, NoOverwrite
	mvi a, CPM_CONIN ; do direct console to A reg so dont effect data at 5Ch & 6Ch FCB's
	MMB 0
	cpi 'n'
	cz ProgExit
	cpi 'N'
	cz ProgExit
	DoDelete:
	lxi d,FCBDEST
	call delete ; remove if present
 	lxi d,FCBDEST
	call make ; create/overwrite the file
	lxi d,nodir ; ready message
	inr a ; 255 becomes 0
	cz ProgExit ; done if no dir space
	; source file open, dest file open
	; copy until end of file on source
	copy: lxi d,FCBSrc ; source
	call read ; read next record
	ora a ; end of file?
	jnz eofile ; skip write if so
	; not end of file, write the record
	lxi d,FCBDEST ; destination
	call write ; write the record
	lxi d,space ; ready message
	ora a ; 00 if write ok
	cnz ProgExit ; end if so
	jmp copy ; loop until eof
	eofile: ; end of file, close destination
	lxi d,FCBDEST ; destination
	call close ; 255 if error
	lxi h,wrprot ; ready message
	inr a ; 255 becomes 00
	cz ProgExit ; shouldn't happen
	; copy operation complete, end
	lxi d,normal ; ready message and fall through to exit msg
	
	ProgExit: ; write message given in de, reboot
	mvi c,printf
	call BDOS ; write message
	mvi c,0
	jmp BDOS ; exit thru reboot system
	
	; =============== Support Routines ================
 ;
 ; system interface subroutines
 ; (all return directly from BDOS)
 ;
	open: mvi c,openf
	jmp BDOS
 ;
	close: mvi c,closef
	jmp BDOS
 ;
	delete: mvi c,deletef
	jmp BDOS
	
	read: mvi c,readf
	jmp BDOS
 ;
	write: mvi c,writef
	jmp BDOS
 ;
	make: mvi c,makef
	jmp BDOS
;


	
	DoExport:
	; source file is open on entry 
	; send to SDA: until end of file on source
	lda ExFileType
	mov c, a ; C Reg is type
	mvi a, CPM_OPEN_EXPORT
	MMB 0 ; open file
	cpi 0FFH ; is it error
	jnz ExpLoop
	lxi d, MMBasicFileErr
	call ProgExit
	ExpLoop: 
	lxi d,FCBSrc ; source
	call read ; read next record
	ora a ; end of file?
	jnz ExpEOF ; skip write to MMBasic file as EOF
	; not end of file, write the record
	mvi a, CPM_TRANSFER_EXPORT
	MMB 0 ; transfer 128 bytes in DMA at 80H
	ora a ; FF if write completed in text file (detected CTRL Z, 1Ah)
	cnz ExpEOF ; end if so
	jmp ExpLoop ; loop until eof
	ExpEOF: ; end of file, close destination
	; copy operation to MMBasic SD card complete, end
	mvi a,CPM_CLOSE_EXPORT
	MMB 0 ; close file
	lda ExFileType
	cpi 1
	jnz ExpEOF_1
	lxi d,ExportTXTComplete ; ready message and fall through
	call ProgExit
	ExpEOF_1:
	lxi d,ExportBINComplete ; ready message and fall through
	call ProgExit
	
	ValidName:
	; ensure valid file names entered
	; assumes HL pointing to start of FCB 
	inx h ; now start of File name
	lxi d,MissingFile
	mov a,m
	cpi ' '  ; missing file name
	cz ProgExit
	mvi c,11 ; 11 bytes to check
	lxi d,FNameError ; ready message
	VNLoop:
	mov a,m
	cpi '?'
	cz ProgExit ; file name invalid
	cpi '*'
	cz ProgExit ; file name invalid
	cpi ':'
	cz ProgExit ; file name invalid, should not have ':' when checking
	inx h
	dcr c
	jnz VNLoop
	ret ; all OK
	
	ValidSrcDest:
	; ensures source and dest are not the same, will exit if so with msg
	lxi h, FCB1+1
	lxi d, FCB2+1
	mvi c,11
	VSDLoop:
	mov b,m 
	ldax d
	cmp b
	rnz ; their different so exit OK
	inx h
	inx d	
	dcr c
	jnz VSDLoop ; check full names including types
	; if we get here we have same names
	lxi d,SameNames
	cz ProgExit ; file name invalid
	
	IsExport:
	; check if destination 80h is SDA: followed by file, ExportFlag set TRUE if PUN: out else set FALSE
	; if so it will get the source file name from the command line at 80h
	lxi h, 80h
	LoopSpace1:
	inx h ; first iteration past lenghth
	mov a,m
	cpi ' '
	jz LoopSpace1 ; skip spaces
	cpi 'S'
	jnz NotExp
	inx h
	mov a,m
	cpi 'D'
	jnz NotExp
	inx h
	mov a,m
	cpi 'A'
	jnz NotExp
	inx h
	mov a,m
	cpi ':'
	jnz NotExp
	; must be export here
	mvi a, TRUE
	sta ExportFlag
	; copy second file name that resides in command line from this point on 
	; here we are at char ':'
	ExLoop1: ; skip spaces
	inx h ; firt iteration past ':'
	mov a,m
	cpi ' '
	jz ExLoop1 ; skip space
	; HL pointing at the first char of the source file, place in 6D till '.'
	mvi b,11
	lxi d, 06dh ; fill location with 11 spaces
	mvi a, ' '
	ExSpLoop:
	stax d
	inx d
	dcr b
	jnz ExSpLoop
	lxi d, 06dh
	mvi b,11
	ExLoop4:
	mov a, m
	cpi '.'
	jz ExLabe5 ; found '.' 
	stax d ; save
	inx h
	inx d
	dcr b
	jz 	ExLabe5 ; did not find ., will get picked up later
	jmp ExLoop4
	ExLabe5:
	; at '.' in command line
	inx h ; skip '.'
	lxi d, 75h ; at where type goes, do for 3 letters
	mov a,m
	stax d
	mov b, a ; save
	inx h
	inx d
	mov a,m
	stax d
	mov c, a ; save
	inx h
	inx d
	mov a,m
	stax d
	inx h
	inx d
	; sort out type, if TXT, BAS, HEX, PRN or ASM then will tell MMBasic its txt else its binary
	;A = last, B=First, C= Second char
	xra b
	xra c
	cpi 'X' ;TXT
	jz ItsText
	cpi 'P' ;BAS
	jz ItsText
	cpi 'U' ; HEX
	jz ItsText
	cpi 'L' ; PRN
	jz ItsText
	cpi '_' ; ASM
	jz ItsText
	NotExp:
	xra a
	sta ExportFlag
	ret	
	ItsText:
	mvi a, 1
	sta ExFileType ; set file type to text
	ret ; done
	
	
	;=============== Local Storage Area =========================
			
	; console messages
	nofile: db 'ERROR no source file$'
	nodir: db 'ERROR no directory space$'
	space: db 'ERROR out of dat space$'
	wrprot: db 'ERROR write protected?$'
	normal: db 'Copy completed successfully$'
	FNameError: db 'ERROR File Name AMBIGUOUS$, e.g use of ? or * in file name$'
	MissingFile: db 'ERROR File Name missing$'
	SameNames: db 'ERROR Source and Destination File Names are the Same$'
	ExportTXTComplete: db 'TEXT File Exported$'
	ExportBINComplete: db 'BINARY File Exported$'
	MMBasicFileErr: db 'MMBASIC File Open Error, Check SD Card$'
	FileExistMSG: db '*** WARNING FILE EXISTS, OVERWRITE? Y/N',13,10,'$'
	NoOverwrite: db 'Aborted Copy$'
	
 ;
 ; data areas
	FCBSrc: ds 33 ; source fcb
	dSrccbcr: equ  FCBSrc+32;
	ExportFlag: ds 1 ; TRUE if exporting 
	ExFileType: db 1 ; set to default txt
 ;
	ds 32 ; 16 level stack
 stack:
	end
 
