/* 
	Colour Maximite 2 Z80 Simulator & CP/M 2.2 OS manager. 
	Version 1.0. August 2022
	Author: GerrL (Backshed)
	Notes:
	1. HLT instruction will return to MMBasic program
	2. IN r (C) instruction, Reg B is NOT Used, normally Reg C and B are used to select port, only Reg C is used.
	3. All IN and OUT codes are actually pin in and out.
	4. INI, INIR, IND, INDR,OUTI, OTIR, OUTD, OTDR  is not supported as it stores bytes and increments address lines, which is not relevant to CMM2.
	5. NMI and INT are LOW active
	6. Mode 0 interrupt will ONLY accept RST (single Byte command)  or CALL instruction only (3 byte command), So only
		set IntMode_0_Code to these commands. Program will not check if the RST or Call have valid routines so if not set will
		cause the program to get lost.
	7. There is some conflict in the docs on when the Aux Carry/Half carry is set with logical operations,
		For AND it is set and for OR and XOR it is reset.
	8. It is up to the user to ensure any assembly files revert back to the simulator monitor using,
			LD A,EXIT_TO_MMBASIC 
			MMB 0		
 */


#include "ARMCFunctions.h" // Version 1.5 used
#include <stdint.h>// only needed for int definitions, e.g uint16_t etc



// ====================================================================
// Compile Options, comment out if not needed
//#define COUNT_INSTRUCT_WHILE_TIMING // For DEbugging this code, will count how many instructions undertaken with DDIXHandler case 18 & 19
#define USE_SUB_FOR_COMPARE // compare uses the stndard subtract function and not 'C' compare
 // if OVERWRITE_GET_ADDRESS is defined then each instruction function will have its get  function address code overwriten once obtained
//  with an unconditional branch. Just speeds up function a little. On speed check there is about a 6% to 10% speed increase so marginal
#define OVERWRITE_GET_ADDRESS 
#define JUMP_18 1 // 18 byte jump for above
#define JUMP_20 2 // 20 byte jump for above
#define MSEC_COUNT_INCLUDED // if defined and COUNT_FLAG_ADDRESS set will count milliseconds,
//#define MM_DEBUG_PRINT_VALUE // uncomment if need DB_PrintByte or DB_PrintInt
// ====================================================================

// ====================================================================
// Specific Addresses used in 64k Memory
#define CPM_START 0x0FA00 // BIOS ENTRY address on RE-START of CP/M
#define DEBUG_FLAG 0xFFFF // set to 44 ('D') if CTRL C to revert to STEP MODE in CP/M and not load CCP
// If MSEC_COUNT_INCLUDED defined
#define COUNT_FLAG_ADDRESS 0x0FFFE // single byte 
#define COUNTER32_ADDRESS 0x0FFF9 // 32 bits or 4 bytes
// ====================================================================

// ====================================================================
// MMB 0 Instruction: A Register value for instruction
// CFunction calls, refer to ARMCFunctions.h
#define CF_USEC 0x0
#define CF_PUT_CONSOLE 0x1
#define CF_GET_CONSOLE 0x2
#define CF_EXT_CFG 0x3
#define CF_EXT_SET 0x4
#define CF_EXT_INP 0x5
#define CF_PIN_SET 0x6
#define CF_PIN_READ 0x7
#define CF_MM_PRINT 0x8
#define CF_INT_TO_STR 0x9
#define CF_CHECK_ABORT 0xA
#define CF_DRAW_RECT 0xB
#define CF_DRAW_LINE 0xC
#define CF_H_RES 0xD
#define CF_V_RES 0xE
#define CF_RUN_MMB_SUB 0xF
#define CF_DRAW_PIXEL 0x10
#define CF_USEC_TIMER 0x11
#define CF_FAST_TIMER 0x12
#define CF_TICK_PER_SEC 0x13
#define CF_DRAW_CIRCLE 0x14
#define CF_DRAW_TRIANGLE 0x15
// CP/M BIOS CALLs
#define CPM_COLD_BOOT 0x20 // Not Used
#define CPM_WARM_BOOT 0x21 // Reloads CCP.Hex or if 0xffff = 0x44 then just reverts to STEP MODE
#define CPM_CONST 0x22 // console status
#define CPM_CONIN 0x23 // Console In to A Reg
#define CPM_CONOUT 0x24 // Console Out, from C Reg
#define CPM_LIST_OUT 0x25 // Write List Out, to Console from C Reg
#define CPM_PUNCH_OUT 0x26 // Not Supported. Write Punch Out, from C Reg
#define CPM_READER 0x27 // Reader In from Console to A Reg
#define CPM_HOME 0x28 // Disk home
#define CPM_SEL_DISK 0x29 // Reg C contains the disk number to select, A=0, B=2, C=3, D=4 (max 4 disks)
#define CPM_MOVE_TO_TRACK 0x2A// C Reg contains track
#define CPM_SETSEC 0x2B // C Reg contains sector
#define CPM_READ_DISK 0x02C // read 128 bytes
#define CPM_WRITE_DISK 0x02D // write 128 bytes
#define CPM_SET_DMA 0x02E // buffer address for disk write and read
// additional CP/M Calls specific for CMM2 
#define CPM_OPEN_EXPORT 0x80 // opens MMBasic file CPM_EXPORT 
#define CPM_TRANSFER_EXPORT 0x81 // transfers 128 byes
#define CPM_CLOSE_EXPORT 0x82 // closes CPM_EXPORT 
// Special handy user functions
#define EXIT_TO_MMBASIC 0x90 // Revert back to the simulator monitor.
#define SET_INT_TIME 0x91 // HL = required microseconds, By default interrupts are checked every 1000 microseconds, this can be changed using this 
#define START_TIMER 0x92 //captures uSecTimer and stores in pReg->nElapseTimer (64 bit)
#define STOP_TIMER 0x93 // Captures uSecTimer and subtracts pReg->nElapseTimer to give elapsed time since Start Timer call. 
#define PC_DEBUG_CNT 0x94 // only used for debugging this program, see switch case for details 
#define MMB_PAGE_DISPLAY 0x95 // calls CMM2 PAGE DISPLAY n with C Reg set to required PAGE
#define MMB_PAGE_WRITE 0x96 // calls CMM2 PAGE WRITE n with C Reg set to required PAGE
#define MMB_CLR 0x97 // Calls MCMM2 CLS
#define MMB_TEXT_OUT 0x98 // Calls CMM2 TEXT, 'HL pointing to memory location laid out as,
							// Byte 0 & 1 = x, Byte 2 & 3 = y, Byte 4 = Alignment ('L', 'C', 'R')
							// Byte 5 =  'T', 'M', 'B', Byte 6 = orientation ('N', 'V', 'I','U','D' 
							// Byte 7 = Font, Byte 8 = Scale, Next 3 bytes for Colour and next 3 bytes for background colour. 
							// Register DE points to a String terminated in 0
#define MMB_GET_TIME 0x99 // Calls CMM2 TIME$, returns with C Reg = seconds, B Reg = Mins, E Reg = Hrs
#define DEBUG_PRINT_CREG 0x9A // Can be called by assembly to print C Register as "C= XX"
#define DEBUG_PRINT_BCREG 0x9B // Can be called by assembly to print BC Register as "BC= XX"
#define MMB_TRACE_REG_STATUS 0x9C // will get MMBasic to send the Z80 status to console
#define MMB_SAVE_IMAGE 0x9D // saves screen to "Z80Prog.bmp"
// Z80 specific
#define Z80_SET_MODE_0_COMMAND 0x40 // only used by Z80 to set IntMode_0_Code[3]

// ====================================================================
//MMBasic Progam Control Flag codes
#define FLAG_STEP 1 // STEP MODE
#define FLAG_END_PROGRAM 2 //No used
#define FLAG_RUN 4 // not used
#define FLAG_INIT 8 // if bit 0 sets any initial data, sets it to 1 after intiating data and address array
#define FLAG_RUN_TO 16 // RUN TO MODE if bit set
#define FLAG_EXIT 32 // If bit Set will exit back to MMBasic

// ====================================================================
// Flags Register Bits
#define FLAG_SIGN 0x80
#define FLAG_ZERO 0x40
#define FLAG_HCARRY 0x10
#define FLAG_PV 0x04
#define FLAG_N 0x02
#define FLAG_CARRY 0x01
#define FLAG_H_AND_N_RESET (~FLAG_HCARRY & ~FLAG_N) // used in an AND to set reset bits

// ====================================================================
// IFF Flags
#define IFF1_BIT 0x01
#define IFF2_BIT 0x04
#define INT_MODE_0_RESET_MASK 0b00111111 // & it with IFF to set to zero
#define INT_MODE_1_SET 0b01000000 // OR with IFF to set to Mode 1
#define INT_MODE_2_SET 0b10000000 // OR with IFF to set to Mode 2

// ====================================================================
// Indicates if Carry Flag used in instructions
#define NO_CARRY 0
#define INCLUDE_CARRY 1

// ====================================================================
// background msec counter  
#define COUNT_FLAG_NOW 1
#define COUNT_FLAG_ELAPSE 2

// ====================================================================
// Key Board buffer size
#define KB_BUFF_SIZE 12 //bytes


#ifdef MM_DEBUG_PRINT_VALUE
void DB_PrintByte(uint8_t dnumber, char cLabel){
// dnumber is the number to print
// cLabel is a single char to define the number, e.g. A = xxxxx
   char cstr[30];
   putConsole(cLabel);
   putConsole('=');
   putConsole(' ');
   IntToStr(cstr,dnumber,10); 
   MMPrintString(cstr);
   putConsole(' ');
   putConsole('\r');
   putConsole('\n');
}


void DB_PrintInt(uint32_t dnumber, char cLabel){
// dnumber is the number to print
// cLabel is a single char to define the number, e.g. A = xxxxx
   char cstr[30];
   putConsole(cLabel);
   putConsole('=');
   putConsole(' ');
   IntToStr(cstr,dnumber,10); 
   MMPrintString(cstr);
   putConsole(' ');
   putConsole('\r');
   putConsole('\n');
}

void DelaySec() {
	int i;
	for(i=0;i<100;i++){
		uSec(10000);
	}
}
#endif

// Determines where in RegUnion
#define RegA 0
#define RegFlag 1
#define RegC 0
#define RegB 1 
#define RegE 0
#define RegD 1
#define RegL 0
#define RegH 1
#define LSB 0
#define MSB 1


// Word Regs (A Reg and Flags,BC,DE, HL, Stack,IY,IX, AltRegs) 
union RegUnion {
   uint16_t WordReg;
   uint8_t RegByte[2];
}; 

struct __attribute__((__packed__)) Registers { // ensure no padding by compiler 
//struct Registers {
	// Z80 Regs
   union RegUnion BCReg; //0&1, C is first 
   union RegUnion DEReg; //2&3, E is first
   union RegUnion HLReg; //4&5, L is first
   union RegUnion AFlagReg; //6&7, A is first , so A = c+6
   union RegUnion StackPtr; //8&9
   uint16_t PC; // byte 10&11
   union RegUnion IX; // byte 12 & 13
   union RegUnion IY; // byte 14 & 15
   union RegUnion AltBCReg; //16 & 17, C is first
   union RegUnion AltDEReg; //18 & 19, E is first
   union RegUnion AltHLReg; //20 & 21, L is first
   union RegUnion AltAFlagReg; //22 & 23, A is first
   uint8_t IReg; // byte 24 used for mode 2 int
   uint8_t IFF; // byte 25, IFF1 bit 1, IFF2, bit 4 (interrupt EI Flip Flops), bit 7 and 8 are the Interupt Mode
   uint16_t MMBParameter; // 26 & 27, this is used by MMBasic for data transfer
   uint8_t nError; // 28, if error occurs set this to an error code, not used
   uint8_t nDebugstep;// 29 how many instruction to process before returning to MMBASIC
   uint16_t nRunToAddress; // 30 & 31, address where to stop running and return to MMBasic program 
   uint8_t ProgCtrlFlags; // 32, used to control program
   char cMDSK[8]; // 33 to 40, name of MMBASIC function that can be called
    // Interrupt Vars
   uint8_t nNMIPin; // 41 The pin number for the NMI interrupt
   uint8_t nINTPin; //42, The pin number for the INT interrupt
   uint8_t IntMode_2_Offset; // byte 43, single byte offset normally supplied by int device but user can supply here.
   uint16_t nIntTime; // 44 & 45, time between ints  
   uint8_t IntMode_0_Code[3]; // 46 to 48, up to 3 byte command normally supplied by int device but user can supply here.
     // CP/M   
   uint16_t nDMAAddress; // 49 & 50, used by CP/M
   char KbBuffer[12]; //byte 51 to 62 = 12 bytes
   uint8_t KBHead; // byte 63, 
   uint8_t KBTail; // byte 64,
   uint8_t nFuncInit; // 65, only used at initilaisation of CSub
   uint16_t nUserVar_66_16; // 66 & 67 these are only used when COUNT_INSTRUCT_WHILE_TIMING id defined
   uint16_t nUserVar_68_16; //68 & 69
   uint8_t nUserVar_70_8; // 70
   uint8_t nFileExportFlag; // 71 used in PUNCH (file export), MMBASIC resets this in intitialise
   uint8_t LogOn; //72 flag
   uint8_t LastKeyPress; //73 used for CTRL Q checks
   uint8_t unused3; //74
   uint64_t nElapseTimer; // Byte 75 to 82, stores timer for elapsed time measurments
    };

void GetRetAdd(unsigned int* padd)
{
// get return address
	*padd = (unsigned int)  __builtin_return_address (0);
}

#ifdef OVERWRITE_GET_ADDRESS
void ByPassGetAddress(unsigned int* padd, int nOffset){
	// will overwrite an unconditonal branch (E0) to jump 18 or 20 bytes to bypass the get address once it is obtained on initialisation
	// if nOffset == JUMP_20 then use 08 (20 bytes) as skip else if JUMP_18 use 07 (18 bytes) 
	uint8_t* n8address = (uint8_t*) *padd;
	n8address +=15; // where to place the branch
	if(nOffset == JUMP_20){
		*n8address = 8; // 8 << 1 = 16 + 4 bytes = 20 bytes (14hex)
	}
	else if (nOffset == JUMP_18) {
		*n8address = 7; // 7 << 1 = 14 + 4 bytes = 18 bytes (12hex)
	}
	n8address++;
	*n8address = 0xE0;// uncon branch
}
#endif



void SetHalfCarryFlag(struct Registers* pReg, uint8_t* pReg1, uint8_t nReg2) { //, uint8_t nAdd) {
	// this is an add check so if comparing or subtracting then Reg2 = negate +1 before calling 
	if( ((*pReg1 & 0x0F) + (nReg2 & 0x0F)) & 0x10){
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
	}
	else {
		pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
	}
	
}

void SetParity(struct Registers* pReg, uint8_t cReg) {
	// from bit twiddling web site
	uint8_t nreg = cReg;
	nreg ^= nreg >> 4;
	nreg &= 0xf;
	nreg = (0x6996 >> nreg) & 1;
	// nreg is 1 if odd, else 0 if even 
	if(nreg) {
		pReg->AFlagReg.RegByte[RegFlag] &=  ~FLAG_PV;// set odd
	}
	else{
		pReg->AFlagReg.RegByte[RegFlag] |=  FLAG_PV;
	}
}

void SetSignZeroFlag(struct Registers* pReg, uint8_t cReg) {
	// only effects FLAG_ZERO & FLAG_SIGN
	if(cReg ==0){
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; 
	}
	else{
		pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_ZERO;
	}
	pReg->AFlagReg.RegByte[RegFlag] &= 0x7F; // reset sign
	pReg->AFlagReg.RegByte[RegFlag] |= cReg & 0x80;
}

void Add_8Bit(struct Registers* pReg, uint8_t* pResultReg, uint8_t nAugEnd, uint8_t IncludeCarry  ){
	// does an add, result in pResultReg
	// all flags impacted
	// if IncludeCarry =1 then factor in entry carry flag, IncludeCarry = 0 then ignore entry carry flag
	// to do a Subtract negate the nAugEnd and add 1 prior to entry
	uint8_t ncarry =0;
	uint8_t nNumber = *pResultReg;
	if(IncludeCarry & (pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY) ){
		ncarry =1;
	}
	pReg->AFlagReg.RegByte[RegFlag] =0; // reset all
	SetHalfCarryFlag(pReg, pResultReg, nAugEnd);
	*pResultReg += (nAugEnd + ncarry);
	// when adding operands with similar signs and the result contains a different sign, the Overflow Flag is set
	// check for unequal signs and flip result with ^0x80 so only do if similar signs
	if( ((nNumber ^ nAugEnd) ^ 0x80) & 0x80){
		// now check answer for unequal sign from original number
		if(((*pResultReg ^ nNumber) & 0x80) != 0) {
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set flag as result sign not the same
		}
	}
	//if(nNumber > (0xFF - nAugEnd) ){
	if ( ((uint16_t)nNumber + (uint16_t) nAugEnd +  (uint16_t) ncarry) > 0xFF){
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
	}
	SetSignZeroFlag(pReg, *pResultReg);

}

void Sub_8Bit(struct Registers* pReg, uint8_t* pResultReg, uint8_t nSubTrahend, uint8_t IncludeCarry ){
	// does a sub, result in pResultReg
	// all flags impacted
	// if IncludeCarry =1 then factor in entry carry flag, IncludeCarry = 0 then ignore entry carry flag
	uint8_t tempbyte = *pResultReg;
	uint8_t ncarry = 0;
	if(IncludeCarry){
		// need to  subtract C
		// If C=1  then negated C is 0 so not adding BUT as we have not added 1 to compliment effectively answer is less one if entry carry is set
		// C=0 then negated C = 1 but adding 1 makes up for the lack of +1 on compliment so carry has no effect on subtraction
		// toggle the carry flag
		if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
			ncarry = 1;
		}
		pReg->AFlagReg.RegByte[RegFlag] ^= FLAG_CARRY;
		Add_8Bit(pReg, pResultReg, ~nSubTrahend, INCLUDE_CARRY); 
	}
	else { // ignore carry
		Add_8Bit(pReg, pResultReg, (~nSubTrahend)+1, NO_CARRY);
	}
	pReg->AFlagReg.RegByte[RegFlag] |= 	FLAG_N;
	if(IncludeCarry){
		if((nSubTrahend+ncarry) > tempbyte){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else{
			pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_CARRY;
		}
	}
	else{
		if(nSubTrahend > tempbyte){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else{
			pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_CARRY;
		}
	}
	if((tempbyte ^ nSubTrahend)& 0x80){ // must be unequal signs
		if(tempbyte & 0x80){ //A-
			if( !((tempbyte-nSubTrahend) & 0x80)){ // sub is +
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV;
			}
		}
		else{// A +
			if( (tempbyte-nSubTrahend) & 0x80){ // sub is neg
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV;
			}
		}
	}
}



void Inc8BitVar(struct Registers* pReg, uint8_t* pVar) {
	// will increment pVar and set flags as per document;
	pReg->AFlagReg.RegByte[RegFlag] &= 1;// retain carry
	if(*pVar == 0x7F){
		pReg->AFlagReg.RegByte[RegFlag] |=  FLAG_PV;
	}
	if( (*pVar & 0x0F)==0x0F){
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;
	}
	*pVar +=1;
	SetSignZeroFlag(pReg, *pVar);
}

void Dec8BitVar(struct Registers* pReg, uint8_t* pVar) {
	// will decrement pVar and set flags as per document;
	pReg->AFlagReg.RegByte[RegFlag] &= 1;// retain carry
	pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N; 
	if(*pVar == 0x80){
		pReg->AFlagReg.RegByte[RegFlag] |=  FLAG_PV;
	}
	if( (*pVar & 0x0F)==0x0){
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;
	}
	*pVar -=1;
	SetSignZeroFlag(pReg, *pVar);
}

void CheckBit(struct Registers* pReg, uint8_t nRegVal, uint8_t nCode){
	// if bit as specified in ncode bit (can be 0 to 7) in nRegVal is 0 then Z flag is set otherwise its reset
	// Note nCode is bits 1,2&3
	if( (nRegVal >> nCode) & 1){ // is bit 1
		pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_ZERO; // bit is 1 so zero reset
	}
	else{
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; 
	}
	pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;
	pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_N;
}

void SetBit(struct Registers* pReg, uint8_t* pRegVal, uint8_t nCode){
	// bit as specified in code bit (can be 0 to 7) in nRegVal is set 
	// Note nCode is bits 1,2&3
	uint8_t n = 1 << nCode;
	*pRegVal |= n;
	
}

void CMP_ARegHandler(struct Registers* pReg, uint8_t DataToCMP) 
{
	// Compares DataToCMP to AReg
	//The zero flag indicates equality. No carry indicates that the accumulator is greater than the specified byte; a
	//carry indicates that the accumulator IS less than the byte. However, the meaning of the carry flag is reversed
	//when the values have different signs or one of the values is complemented.
	#ifdef USE_SUB_FOR_COMPARE
		uint8_t tempbyte = pReg->AFlagReg.RegByte[RegA]; // dont effect RegA
		Sub_8Bit(pReg, &tempbyte, DataToCMP, NO_CARRY );
	#else // else use the below process (not used)
		pReg->AFlagReg.RegByte[RegFlag]=0;
		if(pReg->AFlagReg.RegByte[RegA] == DataToCMP){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO;
		}
		if(pReg->AFlagReg.RegByte[RegA] < DataToCMP){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN; 
		}
		if( (DataToCMP & 0x0F) > (pReg->AFlagReg.RegByte[RegA] & 0x0F)){
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
			}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
		}
		// when subtracting operands with unlike signs and the result contains a different sign, the Overflow Flag is set
		// check for unequal signs and flip result with ^0x80 so only do if Overflow
		if((pReg->AFlagReg.RegByte[RegA] ^ DataToCMP)& 0x80){ // must be unequal signs
			if(pReg->AFlagReg.RegByte[RegA] & 0x80){ //A-
				if( !((pReg->AFlagReg.RegByte[RegA]-DataToCMP) & 0x80)){ // sub is +
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV;
				}
			}
			else{// A +
				if( (pReg->AFlagReg.RegByte[RegA]-DataToCMP) & 0x80){ // sub is neg
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV;
				}
			}
		}
	#endif
}

void DAAMngr(struct Registers* pReg) {
	if( (pReg->AFlagReg.RegByte[RegFlag] & FLAG_N) ){
		// SUB Type
		switch(pReg->AFlagReg.RegByte[RegFlag] & 0x11){ // switch on C and H
			case 0x0:
				// no carry or H
				//  we dont need to do anything if hi and low below 09 Row 1
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)<=0x80) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)>=0x06) ){ //Row 2
				pReg->AFlagReg.RegByte[RegA] += 0xFA;
				break;
			}
			break;
			case 1:
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)>=0x70) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)<=0x09)){ //Row 3
				pReg->AFlagReg.RegByte[RegA] += 0xA0;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			break;
			case 0x11:
			if( (((pReg->AFlagReg.RegByte[RegA] & 0xF0)==0x60) || ((pReg->AFlagReg.RegByte[RegA] & 0xF0)==0x70))
					&& ((pReg->AFlagReg.RegByte[RegA] & 0x0F)>=0x06)){ //Row 3
				pReg->AFlagReg.RegByte[RegA] += 0x9A;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			break;
		}
	}
	else{
		// ADD type
		switch(pReg->AFlagReg.RegByte[RegFlag] & 0x11){ // switch on C and H
			case 0:
				// no carry or H
				//  we dont need to do anything if hi and low below 09 Row 1
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)<=0x80) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)>=0x0A)){ //Row 2
				pReg->AFlagReg.RegByte[RegA] += 6;
				break;
			}
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)>=0xA0) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)<=0x09)){ //Row 4
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)>=0x90) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)>=0x0A)){ //Row 5
				pReg->AFlagReg.RegByte[RegA] += 0x66;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			break;
			case 0x10:
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)<=0x90) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)<=0x03)){ //Row 3
				pReg->AFlagReg.RegByte[RegA] += 0x06;
				break;
			}
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)>=0xA0) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)<=0x03)){ //Row 6
				pReg->AFlagReg.RegByte[RegA] += 0x66;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			break;
			case 0x01:
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)<=0x20) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)<=0x09)){ //Row 7
				pReg->AFlagReg.RegByte[RegA] += 0x60;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)<=0x20) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)>=0x0A)){ //Row 8
				pReg->AFlagReg.RegByte[RegA] += 0x66;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			case 0x11:
			if( ((pReg->AFlagReg.RegByte[RegA] & 0xF0)<=0x30) && ((pReg->AFlagReg.RegByte[RegA] & 0x0F)<=0x03)){ //Row 9
				pReg->AFlagReg.RegByte[RegA] += 0x66;
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				break;
			}
			break;
		}
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
	
}

void Mode_0_INTHandler(struct Registers* pReg, uint8_t* pMemory)
{
	// pushes the contents of the program counter onto the stack to provide a return address and then 
	// carries out command at IntMode_0_Code.
	// Will only accept RST and call instructions
	// Does not check if correct vectors set, if not RST or CALL will ignore Int as PC remains as is
	if( (pReg->IntMode_0_Code[0] & 0b11000111) == 0b11000111) { // its a RST
		pReg->StackPtr.WordReg -=2;
		*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC;
		pReg->PC = (uint16_t )(pReg->IntMode_0_Code[0] & 0x38); // leave bits 3,4 &5	
	}
	else{
		if(pReg->IntMode_0_Code[0] == 0xCD) { //its a call
			pReg->StackPtr.WordReg -=2;
			*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC;
			pReg->PC = pReg->IntMode_0_Code[1];
			pReg->PC |= (pReg->IntMode_0_Code[2])<<8;
		}
	}// else
}


// Z80 0xDD code & extended 8085 for calling ARMCFunction.h calls
// Enter with PC pointing at byte after DD
void DDIXHandler(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory)
{
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t* pInt16; 
		uint32_t* pInt32;
		uint64_t* pInt64;
		uint8_t* pInt8;
		uint32_t ncolour;
		uint32_t nfill;
		char* pchar;
		char nchar;
		char cstr[20];
		//uint8_t ncode = (uint8_t) *(pMemory + pReg->PC++);
		
		OpCode = *(pMemory + pReg->PC);  // get next code after DD
		if(OpCode ==0){ // This is op code DO00, Unused by Z80, this is the extended calls
				// assumes A reg contains the call number, 
				// Refer to each function for what Registers are used 
							
				switch(pReg->AFlagReg.RegByte[RegA]){
					case CF_USEC: //uSec, HL pointing to 16 bit parameter
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						uSec(*pInt16);
					break;
					case CF_PUT_CONSOLE: //putConsole, pReg->BCReg.RegByte[RegC] contains char
						putConsole(pReg->BCReg.RegByte[RegC]);
					break;
					case CF_GET_CONSOLE: //getConsole, Returned in pReg->AFlagReg.RegByte[RegA]
						pReg->AFlagReg.RegByte[RegA] = (char) getConsole();// note int returned 
					break;
					case CF_EXT_CFG: //ExtCfg, HL point to 16bit paramters that are in sequence to how called
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						ExtCfg((uint32_t) *pInt16,(uint32_t) *(pInt16+1),(uint32_t) *(pInt16+2));
					break;
					case CF_EXT_SET: //ExtSet, HL point to 16bit paramters that are in sequence to how called
					// checks if a pin (first param) is a digital output and sets second param as output if it is
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						ExtSet((uint32_t) *pInt16,(uint32_t) *(pInt16+1));
					break;
					case CF_EXT_INP: //ExtInp, C contains pin, HL register has pin data.
						pReg->HLReg.WordReg = (uint16_t) ExtInp((uint32_t) pReg->BCReg.RegByte[RegC]);
					break;
					case CF_PIN_SET: //PinSetBit, Reg C contains pin, B contains second parameter (second bit, refer to ARMCFunctions.h)
						PinSetBit((uint32_t) pReg->BCReg.RegByte[RegC], (uint32_t) pReg->BCReg.RegByte[RegB]);
					break;
					case CF_PIN_READ: //PinRead, Reg C contains pin, HL register has pin data.
						pReg->HLReg.WordReg = (uint16_t) PinRead((uint32_t) pReg->BCReg.RegByte[RegC]);
					break;
					case CF_MM_PRINT: //MMPrintString, DE pointing to 0x00 terminated string
						pchar = (char*) (pMemory + pReg->DEReg.WordReg);
						MMPrintString(pchar);
					break;
					case CF_INT_TO_STR: //IntToStr, DE points to memory string location to be filled in pMemory, 
							  // if C=0 then HL byte offset into registers, if C=1 then HL points to CP/M pMemory	
							 // and B is second parameter base (normally 10 for readability)
						pchar = (char*) (pMemory + pReg->DEReg.WordReg);	
						if(pReg->BCReg.RegByte[RegC]){	 
							pInt32 = (uint32_t*) (pMemory + pReg->HLReg.WordReg);
						}
						else {
							pInt32 = (uint32_t*) (&pReg->BCReg.RegByte[RegC] + (uint32_t)pReg->HLReg.WordReg);
						}
						IntToStr(pchar,*pInt32,(uint32_t) pReg->BCReg.RegByte[RegB]);
					break;
					case CF_CHECK_ABORT: //CheckAbort, returns to command prompt if CTRL Q pressed
							  // Note in CP/M CTRL C returns to warm boot and CTRL Q returns to command prompt
						CheckAbort();
					break;
					//GetMemory, GetTempMemory and FreeMemory cannot be used as we are limited to a fixed 65k which allocated in MMBasic
					case CF_DRAW_RECT: //DrawRectangle, HL points to X1, Y1, X2, Y2, Colour, all params 16 bit except colour which is 24 bits
						pInt8 = (uint8_t*) (pMemory + pReg->HLReg.WordReg) + 8;
						ncolour = *pInt8 + ((*(pInt8+1))<<8) + ((*(pInt8+2)) << 16); //blue + green + red
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						DrawRectangle((uint32_t)*pInt16,(uint32_t)*(pInt16+1),(uint32_t)*(pInt16+2),(uint32_t)*(pInt16+3),ncolour);
					break;
					// Bit map drawing not included 
					case CF_DRAW_LINE: //DrawLine 
						// MMBasic command is LINE x1, y1, x2, y2 [, LW [, C]]
						// HL points to a memory location laid out as little endian WORDS of
						// x1,y1,x2,y2,width,Color (16bits), Colour 8 bits (blue=lsb, then green then red msb), total use 7 words
						pInt8 = (uint8_t*) (pMemory + pReg->HLReg.WordReg) + 10; // pointing at colour (the blue byte)
						ncolour = *pInt8 + ((*(pInt8+1))<<8) + ((*(pInt8+2)) << 16); //blue + green + red
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						DrawLine((uint32_t)*pInt16,(uint32_t)*(pInt16+1),(uint32_t)*(pInt16+2),(uint32_t)*(pInt16+3),(uint32_t)*(pInt16+4),ncolour);
					break;
					//case 0xB: //FontTable
					//	pInt16 = (uint32_t*) (pMemory + pReg->HLReg.WordReg);
					//	*pInt16 = FontTable();
					//break;
					case CF_H_RES: //HRes returned in HL
						pReg->HLReg.WordReg = (uint16_t) HRes;
					break;
					case CF_V_RES: //VRes, returned in HL
						pReg->HLReg.WordReg = (uint16_t) VRes;
					break;
					case CF_RUN_MMB_SUB: //RunBasicSub DE points to name of SUB
						pchar = (char*) (pMemory + pReg->DEReg.WordReg);
						RunBasicSub(pchar); // pchar MUST be terminated with TWO null chars
					break;
					case CF_DRAW_PIXEL: //DrawPixel(x,y,c), X,Y are 16 bit values, colour 24 bit RGB
						// HL points to a memory location laid out as little endian WORDS of
						// x,y and colour
						pInt8 = (uint8_t*) (pMemory + pReg->HLReg.WordReg) + 4;
						ncolour = *pInt8 + ((*(pInt8+1))<<8) + ((*(pInt8+2)) << 16); //blue + green + red
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						DrawPixel((uint32_t)*pInt16,(uint32_t)*(pInt16+1),ncolour); 
					break;
					case CF_USEC_TIMER: //uSecTimer, places value where HL pointing, note its 64 bit (8 bytes)
						pInt64 = (uint64_t*) (pMemory + pReg->HLReg.WordReg);
						*pInt64 = uSecTimer(); 
					break;
					case CF_FAST_TIMER: //FastTimer, places value where HL pointing, note its 64 bit (8 bytes)
						pInt64 = (uint64_t*) (pMemory + pReg->HLReg.WordReg);
						*pInt64 = FastTimer(); 
					break;
					case CF_TICK_PER_SEC: //TicksPerUsec, places value where HL pointing
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						*pInt16 = (uint16_t) TicksPerUsec; 
					break;
					case CF_DRAW_CIRCLE: // DrawCircle(int x, int y, int radius, int w, int c, int fill, MMFLOAT aspect)
						// all params except colour 16 bit, colour 24 bit, aspect is fixed at one
						// HL points to a memory location laid out as little endian WORDS of
						// x,y and colour		
						pInt8 = (uint8_t*) (pMemory + pReg->HLReg.WordReg) + 8;
						ncolour = *pInt8 + ((*(pInt8+1))<<8) + ((*(pInt8+2)) << 16); //blue + green + red
						nfill = *(pInt8+3) + ((*(pInt8+4))<<8) + ((*(pInt8+5)) << 16); //blue + green + red ; 
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						DrawCircle((uint32_t)*pInt16,(uint32_t)*(pInt16+1),(uint32_t)*(pInt16+2),(uint32_t)*(pInt16+3),ncolour,nfill,1.0);  
						
					break;
					case CF_DRAW_TRIANGLE: //DrawTriangle(int x0, int y0, int x1, int y1, int x2, int y2, int c, int fill
							// all params except colours 16 bit, colour 24 bit
							// HL points to a memory location laid out as little endian WORDS of
							// x0, y0, x1, y1, x2, y2, c, fill			
						pInt8 = (uint8_t*) (pMemory + pReg->HLReg.WordReg) + 12;
						ncolour = *pInt8 + ((*(pInt8+1))<<8) + ((*(pInt8+2)) << 16); //blue + green + red
						nfill = *(pInt8+3) + ((*(pInt8+4))<<8) + ((*(pInt8+5)) << 16); //blue + green + red 
						pInt16 = (uint16_t*) (pMemory + pReg->HLReg.WordReg);
						DrawTriangle((uint32_t)*pInt16,(uint32_t)*(pInt16+1),(uint32_t)*(pInt16+2),(uint32_t)*(pInt16+3),(uint32_t)*(pInt16+4),(uint32_t)*(pInt16+5),ncolour,nfill); 
					break;
					
					
					//===============================: CPM CALLS :==========================================
					// Starts at 0x21	
					case CPM_WARM_BOOT: // Warm BOOT, called on CTRL C in CPM
						pReg->MMBParameter=CPM_WARM_BOOT;
						RunBasicSub(pReg->cMDSK);
						break;	
					case CPM_CONST: // KeyBdHit, returns FF in RegA if char in console buffer else 0
						nchar = (char) getConsole();// getconsole return -1 (ff) if no char else will return char off MMBASIC buffer
						if(nchar != 0xFF){ // we have a char, save it cos we are only checking
							if(nchar != 0xA){
								pReg->KbBuffer[pReg->KBTail++] = nchar;
								pReg->LastKeyPress =nchar;
								if(pReg->KBTail == KB_BUFF_SIZE){
									pReg->KBTail = 0;
								}
								pReg->AFlagReg.RegByte[RegA] = 0xFF; // indicate we have char waiting
							}// not LF, dont process line feed
							else{
								pReg->AFlagReg.RegByte[RegA] = 0; // if LF then tell CP/M no char waiting
								pReg->LastKeyPress =0;
							}
						} // have char in MMBasic buffer
						else{ // no console char but check CP/M buffer empty
							pReg->AFlagReg.RegByte[RegA] = (pReg->KBTail == pReg->KBHead) ? 0: 0xFF; // empty if head == tail else FF
						} // else 0
						break;	
					case CPM_CONIN: // The next console character is read into register A, if no char will wait so can hang
						if(pReg->KBTail == pReg->KBHead){ // buffer is empty so loop until char in console
							while(1){
								nchar = (char) getConsole();// getconsole return -1 (ff) if no char
								if(nchar != 0xFF ) { // char in MMBasic buffer
									if(nchar != 0xA){ // dont process LF
										pReg->AFlagReg.RegByte[RegA] = nchar;// & 0x7F;
									return; // have char so end
									} // not LF
								} // have a char
							} // while
						} //  CP/M buffer empty
						else { // we have char in CP/M buffer so use that
							pReg->AFlagReg.RegByte[RegA] = pReg->KbBuffer[pReg->KBHead++];
							if(pReg->KBHead == KB_BUFF_SIZE){
								pReg->KBHead = 0;
							}
						}
					break;		
					case CPM_CONOUT: // The character is sent from register C to the console output device
						putConsole(pReg->BCReg.RegByte[RegC]);
						break;	
					case CPM_LIST_OUT: // The character is sent from register C to the current list output device
						putConsole(pReg->BCReg.RegByte[RegC]);
						break;	
					//case CPM_PUNCH_OUT: // NOT USED
					//		pReg->MMBParameter = CPM_PUNCH_OUT;
					//		RunBasicSub(pReg->cMDSK);
					//		pReg->AFlagReg.RegByte[RegA] = pReg->nFileExportFlag;
					
					//break;	
					case CPM_READER: // The next reader character is read into register A, if no char will wait so can hang
						nchar =-1;
						while(nchar == -1)
							nchar = (char) getConsole();// note int returned 
						break;		
					case CPM_HOME: // Set Disk to start of disk file
						pReg->MMBParameter=CPM_HOME;
						RunBasicSub(pReg->cMDSK);
						break;	
					case CPM_SEL_DISK: // Select the disk, C Reg contains the disk number, A:=0, B: =1, C: =2, D: =3
						// if not available will return pReg->HLReg.WordReg set to 0000 else return FFFF, note the ASM BIOS will set the HL to
						// correct Table if OK
						pReg->MMBParameter=CPM_SEL_DISK;
						RunBasicSub(pReg->cMDSK);
						break;		
					case CPM_MOVE_TO_TRACK: // Reg C contain the track
						pReg->MMBParameter=CPM_MOVE_TO_TRACK;
						RunBasicSub(pReg->cMDSK);
						break;	
					case CPM_SETSEC: //C Reg contains sector
						pReg->MMBParameter=CPM_SETSEC;
						RunBasicSub(pReg->cMDSK);
						break;
					break;
					case CPM_READ_DISK: // A Reg = 0 if OK else 1
						pReg->MMBParameter=CPM_READ_DISK;
						RunBasicSub(pReg->cMDSK);
						break;
					case CPM_WRITE_DISK: // A Reg = 0 if OK else 1
						pReg->MMBParameter=CPM_WRITE_DISK;
						RunBasicSub(pReg->cMDSK);
						break;
					case CPM_SET_DMA: // H&L is address
						pReg->nDMAAddress = pReg->HLReg.WordReg;
						break;
					case Z80_SET_MODE_0_COMMAND: // // only used by Z80 to set IntMode_0_Code[3]
						//C Reg contains the byte to store (e.g. CD or the address bytes)
						//B reg contains the element of IntMode_0_Code array to store into
						if(pReg->BCReg.RegByte[RegB] >=0  && pReg->BCReg.RegByte[RegB]<3){ // ensure with specs
							pReg->IntMode_0_Code[pReg->BCReg.RegByte[RegB]] = pReg->BCReg.RegByte[RegC];
						}
						break;	
					case CPM_OPEN_EXPORT: // opens MMBasic file CPM_EXPORT 
						// The C Reg contains the type, 1 is text, 2 is binary, On return A reg = 0 if OK else FF if error
						
						pReg->MMBParameter=CPM_OPEN_EXPORT;
						RunBasicSub(pReg->cMDSK);
						break;
					case CPM_TRANSFER_EXPORT: // transfers 128 bytes to CPM_EXPORT , will return A reg 0
						// Exports all the Data in the DMA buffer at 0x80 to the SD file, refer to details on the COPY.COM program in manual.
						pReg->MMBParameter=CPM_TRANSFER_EXPORT;
						RunBasicSub(pReg->cMDSK);
						break;
					case CPM_CLOSE_EXPORT: // close CPM_EXPORT 
						pReg->MMBParameter=CPM_CLOSE_EXPORT;
						RunBasicSub(pReg->cMDSK);
						break;	
						
						// ====================== Specials ============================
						case EXIT_TO_MMBASIC: // Will exit CFunction after command executed
						pReg->ProgCtrlFlags |= FLAG_EXIT;
					break;
					case SET_INT_TIME: // Set Interrupt time in case needs to be changed from it default of 1000 microseconds
						// entry is H&L set to microseconds, max is &HFFFE = 65534,
						// If set to FFFF then interrupts not checked
						pReg->nIntTime = pReg->HLReg.WordReg;
					break;
					case START_TIMER: // Start Time, Gets time and stores in nElapseTimer
					#ifdef COUNT_INSTRUCT_WHILE_TIMING //only used in evaluating this simulator
					// TEST ONLY ---------------------						
					pReg->nUserVar_24_16 =0;
					pReg->nUserVar_26_16=0;
					pReg->nUserVar_28_8 =1;
					// ---------------------------------
					#endif
						pReg->nElapseTimer = uSecTimer();
					break;
					case STOP_TIMER: // Elapsed time, Gets time and subtracts pReg->nElapseTimer & stores results in pReg->nElapseTimer
						pReg->nElapseTimer = uSecTimer()- pReg->nElapseTimer;
						#ifdef COUNT_INSTRUCT_WHILE_TIMING
						// TEST ONLY ---------------------							
						pReg->nUserVar_28_8 =0;	// reset so not repeated
						//----------------------------	
						#endif
					break;
					case PC_DEBUG_CNT: // used for DEBUG checking that PC is incrementing correctly
					// HL Reg is set to what the program indicates the PC should be at entry (+1 compared to HL)
						// AReg set to 0 if OK else set to 1
						if(pReg->HLReg.WordReg ==  (pReg->PC-1))
							pReg->AFlagReg.RegByte[RegA] = 0;// good
						else
							pReg->AFlagReg.RegByte[RegA] = 1;// error
					break;	
					case MMB_PAGE_DISPLAY: // C Reg set to Page Number
						pReg->MMBParameter = MMB_PAGE_DISPLAY;
						RunBasicSub(pReg->cMDSK);
					break;
					case MMB_PAGE_WRITE:
						pReg->MMBParameter = MMB_PAGE_WRITE;
						RunBasicSub(pReg->cMDSK);
					break;
					case MMB_CLR:
						pReg->MMBParameter = MMB_CLR;
						RunBasicSub(pReg->cMDSK);
					break;	
					case MMB_TEXT_OUT:
					// TEXT x, y, string$ ,[,alignment$] [, font] [, scale] ,[, c] [, bc]
					// HL pointing to memory location with 16 bits words for x,y, Byte for Alignment ('L', 'C', 'R')
					// Byte for 'T', 'M', 'B', Byte for orientation ('N', 'V', 'I','U','D'
					// Byte for Font, Byte for Scale, 3 bytes for Colour and 3 bytes for background colour. 
					// DE points to a C type String with 0 termination
						pReg->MMBParameter = MMB_TEXT_OUT;
						RunBasicSub(pReg->cMDSK);
					break;
					case MMB_GET_TIME:
					// Return RTC time from the MMBASIC TIME$ function after converting the chars to numbers
					// Returns as C Reg = sec, B Reg = Mins, E Reg = Hrs
						pReg->MMBParameter = MMB_GET_TIME;
						RunBasicSub(pReg->cMDSK);
					break;
					case MMB_SAVE_IMAGE:
						pReg->MMBParameter = MMB_SAVE_IMAGE;
						RunBasicSub(pReg->cMDSK);
					break;
					case DEBUG_PRINT_CREG:
					// print C reg to console
					   putConsole('C');
					   putConsole('=');
					   putConsole(' ');
					   IntToStr(cstr,pReg->BCReg.RegByte[RegC],10); 
					   MMPrintString(cstr);
					   putConsole(' ');
					   putConsole('\r');
					   putConsole('\n');
					break;
					case DEBUG_PRINT_BCREG:
					// prints BC to console
					   putConsole('B');
					   putConsole('C');
					   putConsole('=');
					   putConsole(' ');
					   IntToStr(cstr,pReg->BCReg.WordReg,10); 
					   MMPrintString(cstr);
					   putConsole(' ');
					   putConsole('\r');
					   putConsole('\n');
					break;
					
					default:
						//
					break;

				} // inner switch
				pReg->PC++;// skip 00
				return;
		} // code 00
		else{// Z80 stuff	
			uint16_t nadd;
			uint16_t TempWord1;
			uint16_t TempWord2;
			uint8_t nmem;
			uint8_t n8Temp;
			uint8_t* p8mem;
			// after switch will increment PC one byte so any inst that have parameters will need to inc PC param size
			// PC at byte after DD, i.e DD2Ann[xx] at 2A
			pReg->PC++; //PC at first n
			switch(OpCode){ 
				case 0b00101010: // LD IX, (nn), 2A
					nadd = (uint16_t)*(pMemory + pReg->PC++);
					nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
					pReg->IX.RegByte[LSB] = *(pMemory + nadd);
					pReg->IX.RegByte[MSB] = *(pMemory + nadd+1);
					return;
				break;
				case 0b00110110: // LD (IX+d), n, 00110110 , 36h
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = *(pMemory + pReg->PC+1);
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = *(pMemory + pReg->PC+1);
					}
					pReg->PC +=2;
					return;
				break;
				case 0b00100001: //LD IX, nn, 00100001 , 21h  DD21nnnn
					pReg->IX.WordReg = *(pMemory + pReg->PC) | ((*(pMemory + pReg->PC+1))<<8);
					pReg->PC +=2;
					return;
				break;
				case 0b00100010: //LD (nn), IX ,00100010, 22h, 
					nadd = (uint16_t)*(pMemory + pReg->PC++);
					nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
					*(pMemory + nadd) = pReg->IX.RegByte[LSB];
					*(pMemory + nadd+1) = pReg->IX.RegByte[MSB];
					return;
				break;
				case 0b11111001: //LD SP, IX ,11111001 F9h, DDf9
					pReg->StackPtr.WordReg = pReg->IX.WordReg;
					return;
				break;
				case 0b11100101: //PUSH, IX ,11100101 E5h
					pReg->StackPtr.WordReg -=2;
					*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->IX.WordReg;
					return;
				break;
				case 0xE1: //POP, IX ,11100001 E1h
					pReg->IX.WordReg = *((uint16_t*)(pMemory + pReg->StackPtr.WordReg));
					pReg->StackPtr.WordReg +=2;	
					return;					
				break;
				case 0b11100011: //EX (SP), IX, 11100011 E3h
					TempWord1 = pReg->IX.RegByte[LSB];
					TempWord2 = pReg->IX.RegByte[MSB];
					pReg->IX.RegByte[LSB] = *(pMemory + pReg->StackPtr.WordReg);// lo byte
					pReg->IX.RegByte[MSB] = *(pMemory + pReg->StackPtr.WordReg +1); // hi byte
					*(pMemory + pReg->StackPtr.WordReg) = TempWord1; // low byte
					*(pMemory + pReg->StackPtr.WordReg+1) = TempWord2; // hi byte
					return;
				break;
				case 0b10000110: //ADD A, (IX+d) ,10000110 86h
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, NO_CARRY);
				break;
				case 0b10001110: //ADC A, (IX+d) ,10001110 8Eh
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, INCLUDE_CARRY);
				break;
				case 0b10010110: //SUB A, (IX+d) ,10010110 96h
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, NO_CARRY);
				break;
				case 0b10011110: //SBC A, (IX+d) ,10010110 9Eh
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, INCLUDE_CARRY);
				break;
				case 0b10100110: //AND (IX+d) ,10100110 A6h
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;// res H
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					pReg->AFlagReg.RegByte[RegA] &= nmem;
					SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
					SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				break;
				case 0b10110110: //OR (IX+d) ,10110110 B6h
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					pReg->AFlagReg.RegByte[RegA] |= nmem;
					SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
					SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				break;
				case 0b10101110: //XOR (IX+d) ,10101110 AEh
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					pReg->AFlagReg.RegByte[RegA] ^= nmem;
					SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
					SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				break;
				case 0b10111110: //CP (IX+d) ,10111110 BEh
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N;
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
					Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, NO_CARRY );
				break;
				case 0b00110100: //INC (IX+d) ,00110100,34h
					if((*(pMemory + pReg->PC)) & 0x80){
						Inc8BitVar(pReg, (pMemory + (pReg->IX.WordReg +  (0x100 - *(pMemory + pReg->PC))) ));
					}
					else{
						Inc8BitVar(pReg, (pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ));
					}
				break;
				case 0b00110101: //DEC (IX+d) ,00110101,35h
					if((*(pMemory + pReg->PC)) & 0x80){
						Dec8BitVar(pReg, (pMemory + (pReg->IX.WordReg +  (0x100 - *(pMemory + pReg->PC))) ));
					}
					else{
						Dec8BitVar(pReg, (pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ));
					}
				break;
				case 0b00100011: //INC IX ,00100011,23h DD23
					pReg->IX.WordReg++;
					return;
				break;
				case 0b00101011: //DEC IX ,00101011,2Bh
					pReg->IX.WordReg--;
					return;
				break;
				case 0xE9: //JP (IX)
					pReg->PC = pReg->IX.WordReg;
					return;
				break;
				case 0x46: //LD b, (IX+d), DD46nn
					pReg->BCReg.RegByte[RegB] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x56: //LD d, (IX+d), 
					pReg->DEReg.RegByte[RegD] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x66: //LD h, (IX+d), 
					pReg->HLReg.RegByte[RegH] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x4E: //LD c, (IX+d), 
					pReg->BCReg.RegByte[RegC] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x5E: //LD e, (IX+d), 
					pReg->DEReg.RegByte[RegE] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x6E: //LD l, (IX+d), 
					pReg->HLReg.RegByte[RegL] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x7E: //LD a, (IX+d), 
					pReg->AFlagReg.RegByte[RegA] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x70: //LD [IX+d], b
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->BCReg.RegByte[RegB];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->BCReg.RegByte[RegB];
					}
				break;
				case 0x71: //LD [IX+d], c
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->BCReg.RegByte[RegC];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->BCReg.RegByte[RegC];
					}
				break;
				case 0x72: //LD [IX+d], d
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->DEReg.RegByte[RegD];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->DEReg.RegByte[RegD];
					}
				break;
				case 0x73: //LD [IX+d], e
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->DEReg.RegByte[RegE];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->DEReg.RegByte[RegE];
					}
				break;
				case 0x74: //LD [IX+d], h
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->HLReg.RegByte[RegH];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->HLReg.RegByte[RegH];
					}
				break;
				case 0x75: //LD [IX+d], l
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->HLReg.RegByte[RegL];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->HLReg.RegByte[RegL];
					}
				break;
				case 0x77: //LD [IX+d], a
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->AFlagReg.RegByte[RegA];
					}
					else{
						*(pMemory + (pReg->IX.WordReg +  *(pMemory + pReg->PC)) ) = pReg->AFlagReg.RegByte[RegA];
					}
				break;
				case 0xCB: // CBh cases, IX Bits, these are all 4 bytes long so need to inc PC 3 bytes as they return, PC at nn
					switch( *(pMemory + pReg->PC+1)){ // Byte after nn (i.e DDCBnnXX at XX)
						case 0x06: //RLC (IX+d) ,11001011,CBh,06h
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x80){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem <<= 1;
							*p8mem |= (pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY);
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						case 0x16: //RL (IX+d) ,11001011,CBh,16
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							n8Temp = pReg->AFlagReg.RegByte[RegFlag];
							if(*p8mem & 0x80){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem  <<= 1;
							*p8mem |= (n8Temp & FLAG_CARRY);
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						case 0x0E: // RRC (IX+d), 0E
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
								*p8mem |= 0x80;
							}
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						case 0x1E: // RR (IX+d), 1E
							TempWord1 = pReg->AFlagReg.RegByte[RegFlag];
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							if(TempWord1 & FLAG_CARRY) {
								*p8mem |= 0x80;
							}
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						case  0x26: // SLA (IX+d), 26h
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x80){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem <<= 1;
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						case  0x2E: // SRA (IX+d) 2Eh
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							n8Temp = *p8mem;
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							*p8mem |=(n8Temp & 0x80);
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						case  0x3E: // SRL (IX+d) 3Eh
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; // skip nnXX
							return;
						break;
						default: // of DDCBnnXX, PC at byte after CB (i.e. nn)
							nchar = (*(pMemory + pReg->PC+1) >> 3) & 7;// bits code now low bits (01bbb110 -> .....bbb)
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IX.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IX.WordReg + *(pMemory + pReg->PC)) );
							if(*(pMemory + pReg->PC+1) >= 0x46 && *(pMemory + pReg->PC+1)<= 0x7E){ // BIT
								CheckBit(pReg, *p8mem, nchar);
								pReg->PC +=2; 
								return;
							} 
							if(*(pMemory + pReg->PC+1) >= 0x86 && *(pMemory + pReg->PC+1)<= 0xBE){ // RES
								n8Temp =1;
								n8Temp <<= nchar;
								*p8mem &= ~n8Temp;
								pReg->PC +=2; 
								return;
							} 
							if(*(pMemory + pReg->PC+1) >= 0xC6 && *(pMemory + pReg->PC+1)<= 0xFE){ // SET
								n8Temp =1;
								n8Temp <<= nchar;
								*p8mem |= n8Temp;
								pReg->PC +=2;
								return;
							} 
						break;
					} // CB switch
				break; // end CB
				default: // of outer switch, PC pointing byte after DD
					if( (OpCode & 0x09) ==9 ||  OpCode <= 0x39){ //ADD IX, pp, 00PP1001
						// Shift by 3 so mult 2 as 2 bytes per reg
						switch(OpCode){
							case 0x9: //BC
								nmem = 0;
							break;
							case 0x19: // DE
								nmem = 2;
							break;
							case 0x29: //IX
								nmem = 12; // IX is 12 bytes away from C
							break;
							case 0x39: //SP
								nmem = 8;
							break;
							default:
								nmem = 0;
							break;
						} // switch
						TempWord1 = *((uint16_t*)(&pReg->BCReg.RegByte[RegC] + nmem));// get reg
						if( ((pReg->IX.WordReg & 0x07FF) + (TempWord1 & 0x07FF)) > 0x7FF){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
						}
						if( ((uint32_t)pReg->IX.WordReg + (uint32_t)TempWord1) > 0xFFFF){
						//if(pReg->IX.WordReg >= (0xFFFF - TempWord1) ){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						pReg->IX.WordReg += TempWord1;
						pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_N;
						return; // only 2 byte instruction
					} // if
			} // outer switch
			pReg->PC++;//jump byte after second opcode on 3 byte instructions
		} // else Z80
	} // else get address
}

// Enter with PC pointing next byte after FD
void   FDIYHandler(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory)
{
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t nadd;
			uint16_t TempWord1;
			uint16_t TempWord2;
			uint8_t nmem;
			uint8_t n8Temp;
			uint8_t* p8mem;
			char nchar;
			// after switch will increment PC one byte so any inst that have parameters will need to inc PC param size
			// PC at byte after DD, i.e DD2Ann[xx] at 2A
			OpCode = *(pMemory + pReg->PC);
			pReg->PC++; //PC at first n
			switch(OpCode){ 
				case 0x2A: // LD IY, (nn), 2A
					nadd = (uint16_t)*(pMemory + pReg->PC++);
					nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
					pReg->IY.RegByte[LSB] = *(pMemory + nadd);
					pReg->IY.RegByte[MSB] = *(pMemory + nadd+1);
					return;
				break;
				case 0x36: // LD (IY+d), n, 00110110 , 36h
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = *(pMemory + pReg->PC+1);
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = *(pMemory + pReg->PC+1);
					}
					pReg->PC +=2;
					return;
				break;
				case 0x21: //LD IY, nn, 00100001 , 21h  DD21nnnn, PC at n
					pReg->IY.WordReg = *(pMemory + pReg->PC) | ((*(pMemory + pReg->PC+1))<<8);
					pReg->PC +=2;
					return;
				break;
				case 0b00100010: //LD (nn), IY ,00100010, 22h, 
					nadd = (uint16_t)*(pMemory + pReg->PC++);
					nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
					*(pMemory + nadd) = pReg->IY.RegByte[LSB];
					*(pMemory + nadd+1) = pReg->IY.RegByte[MSB];
					return;
				break;
				case 0b11111001: //LD SP, IY ,11111001 F9h, DDf9
					pReg->StackPtr.WordReg = pReg->IY.WordReg;
					return;
				break;
				case 0b11100101: //PUSH, IY ,11100101 E5h
					pReg->StackPtr.WordReg -=2;
					*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->IY.WordReg;
					return;
				break;
				case 0xE1: //POP, IY ,11100001 E1h
					pReg->IY.WordReg = *((uint16_t*)(pMemory + pReg->StackPtr.WordReg));
					pReg->StackPtr.WordReg +=2;	
					return;					
				break;
				case 0b11100011: //EX (SP), IY, 11100011 E3h
					TempWord1 = pReg->IY.RegByte[LSB];
					TempWord2 = pReg->IY.RegByte[MSB];
					pReg->IY.RegByte[LSB] = *(pMemory + pReg->StackPtr.WordReg);// lo byte
					pReg->IY.RegByte[MSB] = *(pMemory + pReg->StackPtr.WordReg +1); // hi byte
					*(pMemory + pReg->StackPtr.WordReg) = TempWord1; // low byte
					*(pMemory + pReg->StackPtr.WordReg+1) = TempWord2; // hi byte
					return;
				break;
				case 0b10000110: //ADD A, (IY+d) ,10000110 86h
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, NO_CARRY);
				break;
				case 0b10001110: //ADC A, (IY+d) ,10001110 8Eh
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, INCLUDE_CARRY);
				break;
				case 0b10010110: //SUB A, (IY+d) ,10010110 96h
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, NO_CARRY);
				break;
				case 0b10011110: //SBC A, (IY+d) ,10010110 9Eh
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, INCLUDE_CARRY);
				break;
				case 0b10100110: //AND (IY+d) ,10100110 A6h
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;// res H
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					pReg->AFlagReg.RegByte[RegA] &= nmem;
					SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
					SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				break;
				case 0b10110110: //OR (IY+d) ,10110110 B6h
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					pReg->AFlagReg.RegByte[RegA] |= nmem;
					SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
					SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				break;
				case 0b10101110: //XOR (IY+d) ,10101110 AEh
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					pReg->AFlagReg.RegByte[RegA] ^= nmem;
					SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
					SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				break;
				case 0b10111110: //CP (IY+d) ,10111110 BEh
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N;
					nmem = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
					Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], nmem, NO_CARRY );
				break;
				case 0b00110100: //INC (IY+d) ,00110100,34h
					if((*(pMemory + pReg->PC)) & 0x80){
						Inc8BitVar(pReg, (pMemory + (pReg->IY.WordReg +  (0x100 - *(pMemory + pReg->PC))) ));
					}
					else{
						Inc8BitVar(pReg, (pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ));
					}
				break;
				case 0b00110101: //DEC (IY+d) ,00110101,35h
					if((*(pMemory + pReg->PC)) & 0x80){
						Dec8BitVar(pReg, (pMemory + (pReg->IY.WordReg +  (0x100 - *(pMemory + pReg->PC))) ));
					}
					else{
						Dec8BitVar(pReg, (pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ));
					}
				break;
				case 0b00100011: //INC IY ,00100011,23h DD23
					pReg->IY.WordReg++;
					return;
				break;
				case 0b00101011: //DEC IY ,00101011,2Bh
					pReg->IY.WordReg--;
					return;
				break;
				case 0xE9: //JP (IY)
					pReg->PC = pReg->IY.WordReg;
					return;
				break;
				case 0x46: //LD b, (IY+d), DD46nn
					pReg->BCReg.RegByte[RegB] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x56: //LD d, (IY+d), 
					pReg->DEReg.RegByte[RegD] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x66: //LD h, (IY+d), 
					pReg->HLReg.RegByte[RegH] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x4E: //LD c, (IY+d), 
					pReg->BCReg.RegByte[RegC] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x5E: //LD e, (IY+d), 
					pReg->DEReg.RegByte[RegE] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x6E: //LD l, (IY+d), 
					pReg->HLReg.RegByte[RegL] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x7E: //LD a, (IY+d), 
					pReg->AFlagReg.RegByte[RegA] = *(pMemory + pReg->PC) & 0x80 ? *(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ):
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) );
				break;
				case 0x70: //LD [IY+d], b
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->BCReg.RegByte[RegB];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->BCReg.RegByte[RegB];
					}
				break;
				case 0x71: //LD [IY+d], c
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->BCReg.RegByte[RegC];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->BCReg.RegByte[RegC];
					}
				break;
				case 0x72: //LD [IY+d], d
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->DEReg.RegByte[RegD];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->DEReg.RegByte[RegD];
					}
				break;
				case 0x73: //LD [IY+d], e
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->DEReg.RegByte[RegE];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->DEReg.RegByte[RegE];
					}
				break;
				case 0x74: //LD [IY+d], h
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->HLReg.RegByte[RegH];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->HLReg.RegByte[RegH];
					}
				break;
				case 0x75: //LD [IY+d], l
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->HLReg.RegByte[RegL];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->HLReg.RegByte[RegL];
					}
				break;
				case 0x77: //LD [IY+d], a
					if((*(pMemory + pReg->PC)) & 0x80){
						*(pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC))) ) = pReg->AFlagReg.RegByte[RegA];
					}
					else{
						*(pMemory + (pReg->IY.WordReg +  *(pMemory + pReg->PC)) ) = pReg->AFlagReg.RegByte[RegA];
					}
				break;
				case 0xCB: // CBh cases, IY Bits, these are all 4 bytes long so need to inc PC 3 bytes as they return, PC at nn
					switch( *(pMemory + pReg->PC+1)){ // Byte after nn (i.e FDCBnnXX at XX)
						case 0x06: //RLC (IY+d) ,11001011,CBh,06h
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x80){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem <<= 1;
							*p8mem |= (pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY);
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; 
							return;
						break;
						case 0x16: //RL (IY+d) ,11001011,CBh,16
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							n8Temp = pReg->AFlagReg.RegByte[RegFlag];
							if(*p8mem & 0x80){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem  <<= 1;
							*p8mem |= (n8Temp & FLAG_CARRY);
							pReg->PC +=2; 
							return;
						break;
						case 0x0E: // RRC (IY+d), 0E
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
								*p8mem |= 0x80;
							}
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; 
							return;
						break;
						case 0x1E: // RR (IY+d), 1E
							TempWord1 = pReg->AFlagReg.RegByte[RegFlag];
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							if(TempWord1 & FLAG_CARRY) {
								*p8mem |= 0x80;
							}
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; 
							return;
						break;
						case  0x26: // SLA (IY+d), 26h
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x80){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem <<= 1;
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; 
							return;
						break;
						case  0x2E: // SRA (IY+d) 2Eh
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							n8Temp = *p8mem;
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							*p8mem |=(n8Temp & 0x80);
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; 
							return;
						break;
						case  0x3E: // SRL (IY+d) 3Eh
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							if(*p8mem & 0x01){
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
							}
							else {
								pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
							}
							*p8mem >>= 1;
							SetSignZeroFlag(pReg, *p8mem);
							SetParity(pReg, *p8mem);
							pReg->PC +=2; 
							return;
						break;
						default: // of DDCBnnXX, PC at byte after CB (i.e. nn)
							nchar = (*(pMemory + pReg->PC+1) >> 3) & 7;// bits code now low bits (01bbb110 -> .....bbb)
							p8mem = *(pMemory + pReg->PC) & 0x80 ? (uint8_t*) (pMemory + (pReg->IY.WordReg -  (0x100 - *(pMemory + pReg->PC)))):
									(uint8_t*) (pMemory + (pReg->IY.WordReg + *(pMemory + pReg->PC)) );
							if(*(pMemory + pReg->PC+1) >= 0x46 && *(pMemory + pReg->PC+1)<= 0x7E){ // BIT
								CheckBit(pReg, *p8mem, nchar);
								pReg->PC +=2; 
								return;
							} 
							if(*(pMemory + pReg->PC+1) >= 0x86 && *(pMemory + pReg->PC+1)<= 0xBE){ // RES
								n8Temp =1;
								n8Temp <<= nchar;
								*p8mem &= ~n8Temp;
								pReg->PC +=2; 
								return;
							} 
							if(*(pMemory + pReg->PC+1) >= 0xC6 && *(pMemory + pReg->PC+1)<= 0xFE){ // SET
								n8Temp =1;
								n8Temp <<= nchar;
								*p8mem |= n8Temp;
								pReg->PC +=2;
								return;
							} 
						break;
					} // CB switch
				break; // end CB
				default: // of outer switch, PC pointing byte after DD
				if( (OpCode & 0x09) ==9 ||  OpCode <= 0x39){ //ADD IY, pp, 00PP1001
						// Shift by 3 so mult 2 as 2 bytes per reg
						switch(OpCode){
							case 0x9: //BC
								nmem = 0;
							break;
							case 0x19: // DE
								nmem = 2;
							break;
							case 0x29: //HL
								nmem = 14; // IY is 14 bytes away from C
							break;
							case 0x39: //SP
								nmem = 8;
							break;
							default:
								nmem = 0;
							break;
						} // switch
						TempWord1 = *((uint16_t*)(&pReg->BCReg.RegByte[RegC] + nmem));// get reg
						//if( ((pReg->IY.WordReg & 0x07FF) + (TempWord1 & 0x07FF)) > 0x7FF){
						if( ((uint32_t)pReg->IY.WordReg + (uint32_t)TempWord1) > 0xFFFF) {	
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
						}
						if(pReg->IY.WordReg >= (0xFFFF - TempWord1) ){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						pReg->IY.WordReg += TempWord1;
						pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_N;
						return; // 2 byte instruction
					} // if
			} // outer switch
			pReg->PC++;//jump byte after DD
	} // else not address
}

// 11101101, ED
//Enters with PC pointing at byte after ED, will inc PC after switch so any inst with params will need 2 incs
void   EDHandler(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t nadd;
		uint8_t n8Temp;
		uint8_t n8Temp2;
		uint8_t* p8Temp;
		uint8_t n8Temp3;
		uint16_t Reg16Val;
		OpCode = *(pMemory + pReg->PC);  // get next code after ED
		switch(OpCode){
			case 0x57: // LD A,I, 01010111, 57
				pReg->AFlagReg.RegByte[RegFlag] &=FLAG_CARRY; // retain carry
				pReg->AFlagReg.RegByte[RegA] = pReg->IReg;
				pReg->AFlagReg.RegByte[RegFlag] |=  (FLAG_PV & (pReg->IFF & IFF2_BIT)); //IFF2 -> parity flag
				SetSignZeroFlag( pReg, pReg->AFlagReg.RegByte[RegA]);// Note FLAG_HCARRY NOT set to IFF2
			break;
			case 0x5F: // LD A, R, 01011111, 5F
				// not supported
				pReg->AFlagReg.RegByte[RegFlag] |=  (FLAG_PV | (pReg->IFF & IFF2_BIT)); //IFF2 -> parity flag
			break;
			case 0x47: // LD I, A, 01000111, 47
				pReg->IReg = pReg->AFlagReg.RegByte[RegA];
			break;
			case 0x4F: // LD R, A, 01001111, 4F
				// not supported
			break;
			case 0xA0: // LDI, 10100000, A0
				*(pMemory + pReg->DEReg.WordReg) = *(pMemory + pReg->HLReg.WordReg); 
				pReg->DEReg.WordReg++;
				pReg->HLReg.WordReg++;
				pReg->BCReg.WordReg--;
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset H
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_N; // reset N
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
			break;
			case 0xB0: // LDIR 10110000, B0
				do{ // note PC is not decremented by 2 as as per document as this is in a do loop so same outcome, int not recognised here
					*(pMemory + pReg->DEReg.WordReg) = *(pMemory + pReg->HLReg.WordReg); 
					pReg->DEReg.WordReg++;
					pReg->HLReg.WordReg++;
					pReg->BCReg.WordReg--;
				} while(pReg->BCReg.WordReg !=0);
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset H
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_N; // reset N
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
			break;
			case 0xA8: // LDD, 10101000, A8
				*(pMemory + pReg->DEReg.WordReg) = *(pMemory + pReg->HLReg.WordReg); 
				pReg->DEReg.WordReg--;
				pReg->HLReg.WordReg--;
				pReg->BCReg.WordReg--;
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset H
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_N; // reset N
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
			break;
			case 0xB8: // LDDR 10111000, B8
				do{ // note PC is not decremented by 2 as as per document as this is in a do loop so same outcome, int not recognised here
					*(pMemory + pReg->DEReg.WordReg) = *(pMemory + pReg->HLReg.WordReg); 
					pReg->DEReg.WordReg--;
					pReg->HLReg.WordReg--;
					pReg->BCReg.WordReg--;
				} while(pReg->BCReg.WordReg !=0);
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset H
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_N; // reset N
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
			break;
			case 0xA1: // CPI, 10100001, A1
				pReg->AFlagReg.RegByte[RegFlag] &= FLAG_CARRY; // retain carry, rest zero
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N; // set
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
				if(pReg->AFlagReg.RegByte[RegA] == *(pMemory + pReg->HLReg.WordReg)){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; // set
				}
				else {
					if(pReg->AFlagReg.RegByte[RegA] < *(pMemory + pReg->HLReg.WordReg)){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN;
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_SIGN; // reset
					}
					SetHalfCarryFlag(pReg, &pReg->AFlagReg.RegByte[RegA], ~(*((uint8_t*)(pMemory + pReg->HLReg.WordReg)))+1);
				} // else neg result
				pReg->HLReg.WordReg++;
				pReg->BCReg.WordReg--;
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
			break;
			case 0xB1: // CPIR, 10110001, B1
				pReg->AFlagReg.RegByte[RegFlag] &= FLAG_CARRY; // retain carry, rest zero
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N; // set
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
				do{
				if(pReg->AFlagReg.RegByte[RegA] == *( pMemory + pReg->HLReg.WordReg)){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; // set
				}
				else {
					if(pReg->AFlagReg.RegByte[RegA] < *(pMemory + pReg->HLReg.WordReg)){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN;
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_SIGN; // reset
					}
				} // else not equal
					pReg->HLReg.WordReg++;
					pReg->BCReg.WordReg--;
				} while( (pReg->BCReg.WordReg !=0) && ( !(pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO)) );
				SetHalfCarryFlag(pReg, &pReg->AFlagReg.RegByte[RegA], ~(*((uint8_t*)(pMemory + pReg->HLReg.WordReg-1)))+1);// only do on exit
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
			break;
			case 0xA9: // CPD, 10101001, A9
				pReg->AFlagReg.RegByte[RegFlag] &= FLAG_CARRY; // retain carry, rest zero
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N; // set
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
				if(pReg->AFlagReg.RegByte[RegA] == *(pMemory + pReg->HLReg.WordReg)){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; // set
				}
				else {
					if(pReg->AFlagReg.RegByte[RegA] < *(pMemory + pReg->HLReg.WordReg)){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN;
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_SIGN; // reset
					}
					SetHalfCarryFlag(pReg, &pReg->AFlagReg.RegByte[RegA], *((uint8_t*)(pMemory + pReg->HLReg.WordReg)));
				} // else neg result
				pReg->HLReg.WordReg--;
				pReg->BCReg.WordReg--;
				if(pReg->BCReg.WordReg !=0){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
				}
			break;
			case 0xB9: // CPDR, 10111001, B9
				pReg->AFlagReg.RegByte[RegFlag] &= FLAG_CARRY; // retain carry, rest zero
				pReg->AFlagReg.RegByte[RegFlag] |= FLAG_N; // set
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
				do{
				if(pReg->AFlagReg.RegByte[RegA] == *( pMemory + pReg->HLReg.WordReg)){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; // set
				}
				else {
					if(pReg->AFlagReg.RegByte[RegA] < *(pMemory + pReg->HLReg.WordReg)){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN;
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_SIGN; // reset
					}
					} // else neg result
				pReg->HLReg.WordReg--;
				pReg->BCReg.WordReg--;
				} while( (pReg->BCReg.WordReg !=0) &&( !(pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO)) );
				SetHalfCarryFlag(pReg, &pReg->AFlagReg.RegByte[RegA], ~(*((uint8_t*)(pMemory + pReg->HLReg.WordReg-1)))+1);// only do on exit
				pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_PV; // reset
			break;
			case 0x44: // NEG, 01000100, 44
				// 0-Areg
				n8Temp =0;
				Sub_8Bit(pReg, &n8Temp, pReg->AFlagReg.RegByte[RegA], NO_CARRY);
				pReg->AFlagReg.RegByte[RegA] = n8Temp;
			break;
			case 0x46: // IM 0, 01000110, 46
				pReg->IFF &= INT_MODE_0_RESET_MASK;
			break;
			case 0x56: // IM 1, 01010110, 56
				pReg->IFF &= INT_MODE_0_RESET_MASK;
				pReg->IFF |= INT_MODE_1_SET;
			break;
			case 0x5E: // IM 2, 01011110, 5E
				pReg->IFF &= INT_MODE_0_RESET_MASK;
				pReg->IFF |= INT_MODE_2_SET;
			break;
			case 0x6F: // RLD, 01101111, 6F
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				n8Temp = *(pMemory + pReg->HLReg.WordReg); // mem original
				n8Temp3 = pReg->AFlagReg.RegByte[RegA] & 0b00001111;
				*(pMemory + pReg->HLReg.WordReg) = 0;
				pReg->AFlagReg.RegByte[RegA] &= 0b11110000; 
				*(pMemory + pReg->HLReg.WordReg) |= n8Temp<<4; // low mem bits to hi mem bits
				pReg->AFlagReg.RegByte[RegA] |= n8Temp>>4; // hi mem bits to A low bits
				*(pMemory + pReg->HLReg.WordReg) |= n8Temp3;
				SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
			break;
			case 0x67: // RRD, 01100111, 67
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				n8Temp = *(pMemory + pReg->HLReg.WordReg); // mem original
				n8Temp3 = pReg->AFlagReg.RegByte[RegA] & 0b00001111;
				*(pMemory + pReg->HLReg.WordReg) = 0;
				pReg->AFlagReg.RegByte[RegA] |= n8Temp & 0b00001111; //low bits mem to low bits A
				*(pMemory + pReg->HLReg.WordReg) |= n8Temp3 << 4; //low bit A to hi bits mem
				*(pMemory + pReg->HLReg.WordReg) |= n8Temp >> 4;
				SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
				SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
			break;
			case 0x4D: // RETI, Ret from maskable int
				// POP off Stack and place in PC
				pReg->PC = *((uint16_t*)(pMemory + pReg->StackPtr.WordReg));
				pReg->StackPtr.WordReg += 2;
			break;
			case 0x45: // RETN, return from NMI
				// POP off Stack and place in PC
				pReg->PC = *((uint16_t*)(pMemory + pReg->StackPtr.WordReg));
				pReg->StackPtr.WordReg += 2;
			break;
			//case 0xA2: // INI Not Supported
			default:
			pReg->PC++; //skip past 2nd instruction code
			switch(OpCode & 0b11001111){// remove registers from switch
				case 0x4B: //LD dd, (nn), 01dd1011, 0x4B is what is left after reg offset removed
					nadd = (uint16_t)*(pMemory + pReg->PC++);//
					nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
					// Shift by 3 so mult 2 as 2 bytes per reg
					n8Temp = (OpCode & 0b00110000) >> 3;
					if(n8Temp == 6){
						n8Temp +=2; // skip Areg & Flags
					}
					*((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp)) = *((uint16_t*)(pMemory + nadd));
					return; 
				break;
				case 0x43: //LD [nn], dd, 01dd0011
					nadd = (uint16_t) (*(pMemory + pReg->PC++));//
					nadd |= (uint16_t) ((*(pMemory + pReg->PC++))<< 8);
					n8Temp = (OpCode & 0b00110000) >> 3;
					if(n8Temp == 6){
						n8Temp +=2; // skip Areg & Flags
					}
					*((uint16_t*)(pMemory + nadd)) = *((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
					return;
				break;
				case 0x4A: // ADC HL,ss, 01SS1010
					// Shift by 3 so mult 2 as 2 bytes per reg
					n8Temp = (OpCode & 0b00110000) >> 3;
					if(n8Temp == 6){
						n8Temp +=2; // skip Areg & Flags
					}
					Reg16Val = *((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
					if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
						Reg16Val++; //add 1 for carry
					}
					pReg->AFlagReg.RegByte[RegFlag] = 0;
					if( ((pReg->HLReg.WordReg & 0x07FF) + (Reg16Val & 0x07FF)) > 0x7FF){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
					}
					if(pReg->HLReg.WordReg >= (0xFFFF - Reg16Val) ){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
					}
					// when adding operands with similar signs and the result contains a different sign, the Overflow Flag is set
					// check for unequal signs and flip result with ^0x8000 so only do if Overflow
					if( ((pReg->HLReg.RegByte[RegH] ^ Reg16Val) ^ 0x8000) & 0x8000){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set flag as result sign not the same
					}
					pReg->HLReg.WordReg += Reg16Val;
					if(pReg->HLReg.WordReg == 0){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; 
					}
					if(pReg->HLReg.WordReg & 0x8000){ // is neg
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN;
					}
				break;
				case 0x42: // SBC HL,ss, 01SS0010
					// Shift by 3 so mult 2 as 2 bytes per reg
					n8Temp2 = pReg->AFlagReg.RegByte[RegFlag]; // save carry
					pReg->AFlagReg.RegByte[RegFlag] = FLAG_N;
					n8Temp = (OpCode & 0b00110000) >> 3;
					if(n8Temp == 6){
						n8Temp +=2; // skip Areg & Flags
					}
					Reg16Val = *((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
					if(n8Temp2 & FLAG_CARRY){
						Reg16Val++; //add 1 for carry, will be adding to subtract
					}
					if( (pReg->HLReg.WordReg & 0x07FF) < (Reg16Val & 0x07FF)){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
					}
					Reg16Val = ~Reg16Val; // set for add
					Reg16Val++; // two compliment
					
					if(pReg->HLReg.WordReg >= (0xFFFF - Reg16Val) ){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
					}
					else {
						pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
					}
					// when adding operands with similar signs and the result contains a different sign, the Overflow Flag is set
					// check for unequal signs and flip result with ^0x8000 so only do if Overflow
					if( ((pReg->HLReg.RegByte[RegH] ^ Reg16Val) ^ 0x8000) & 0x8000){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_PV; // set flag as result sign not the same
					}
					pReg->HLReg.WordReg += Reg16Val;
					if(pReg->HLReg.WordReg == 0){
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_ZERO; 
					}
					if(pReg->HLReg.WordReg & 0x8000){ // is neg
						pReg->AFlagReg.RegByte[RegFlag] |= FLAG_SIGN;
					}
				break;
				default:
					switch(OpCode & 0b11000111){
						case 0b01000000: //IN r (C), 01rrr000, Note Reg B is NOT Used
							pReg->AFlagReg.RegByte[RegFlag] = 0;
							// B 000, C 001, D 010, E 011, H 100, L 101, A 111
							n8Temp2 = ((OpCode & 0b00111000)>> 3);
							if(n8Temp2 != 6){
								p8Temp = (&pReg->BCReg.RegByte[RegC] + n8Temp2);
								*p8Temp = (uint8_t) PinRead((uint32_t) pReg->BCReg.RegByte[RegC]);
								SetSignZeroFlag(pReg, *p8Temp);
								SetParity(pReg, *p8Temp);
							}
						break;
						case 0b01000001: //OUT r (C), 01rrr001, Note Reg B is NOT Used
							// B 000, C 001, D 010, E 011, H 100, L 101, A 111
							n8Temp2 = ((OpCode & 0b00111000)>> 3);
							if(n8Temp2 != 6){
								ExtSet((int) pReg->BCReg.RegByte[RegC], (int)*(&pReg->BCReg.RegByte[RegC] + n8Temp2));
							}
						break;
					} // default default switch
					//return;
			} // inner switch in default
			return;
		} // end outer switch
		pReg->PC++;
	} // not getting address
}

// 11001011, CB, 
// enters with PC pointing at byte after CB. All 2 byte inst so just need to Inc PC once
void   CBHandler(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint8_t* p8mem;
		uint8_t* p8Reg;
		uint8_t n8Temp;
		uint8_t n8Temp2;
		OpCode = *(pMemory + pReg->PC);  // get next code after CB
		switch(OpCode){
			case 06: // RLC (HL), 06
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				if(*p8mem & 0x80){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem <<= 1;
				*p8mem |= (pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY);
				SetSignZeroFlag(pReg, *p8mem);
				SetParity(pReg, *p8mem);
			break;
			case 0x16: // RL (HL), 16
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				n8Temp = pReg->AFlagReg.RegByte[RegFlag];
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				if(*p8mem & 0x80){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem <<= 1;
				*p8mem |= (n8Temp & FLAG_CARRY);
			break;
			case 0x0E: // RRC (HL), 0E
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				n8Temp = pReg->AFlagReg.RegByte[RegFlag];
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				if(*p8mem & 0x01){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem >>= 1;
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
					*p8mem |= 0x80;
				}
				SetSignZeroFlag(pReg, *p8mem);
				SetParity(pReg, *p8mem);
			break;
			case 0x1E: // RR (HL), 1Eh
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				n8Temp = pReg->AFlagReg.RegByte[RegFlag];
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				if(*p8mem & 0x01){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem >>= 1;
				if(n8Temp & FLAG_CARRY) {
					*p8mem |= 0x80;
				}
				SetSignZeroFlag(pReg, *p8mem);
				SetParity(pReg, *p8mem);
			break;
			case  0x26: // SLA (HL), 26h
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				if(*p8mem & 0x80){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem <<= 1;
				SetSignZeroFlag(pReg, *p8mem);
				SetParity(pReg, *p8mem);
			break;
			case  0x2E: // SRA (HL) 2Eh
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				n8Temp = *p8mem;
				if(*p8mem & 0x01){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem >>= 1;
				*p8mem |=(n8Temp & 0x80);
				SetSignZeroFlag(pReg, *p8mem);
				SetParity(pReg, *p8mem);
			break;
			case  0x3E: // SRL (HL) 3Eh
				pReg->AFlagReg.RegByte[RegFlag] = 0;
				p8mem = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
				if(*p8mem & 0x01){
					pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
				}
				else {
					pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
				}
				*p8mem >>= 1;
				SetSignZeroFlag(pReg, *p8mem);
				SetParity(pReg, *p8mem);
			break;
			
			default:
				switch(OpCode & 0b11111000){
					case 0: // RLC r, 00000RRR
						pReg->AFlagReg.RegByte[RegFlag] = 0;
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp =OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						if(*p8Reg & 0x80){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg <<= 1;
						*p8Reg |= (pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY);
						SetSignZeroFlag(pReg, *p8Reg);
						SetParity(pReg, *p8Reg);
					break;
					case  0x10: // RL r, 00010RRR
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp = OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						n8Temp = pReg->AFlagReg.RegByte[RegFlag];
						if(*p8Reg & 0x80){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg <<= 1;
						*p8Reg |= (n8Temp & FLAG_CARRY);
					break;
					case  0x08: // RRC r, 00001RRR
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp = OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						n8Temp = pReg->AFlagReg.RegByte[RegFlag];
						if(*p8Reg & 0x01){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg >>= 1;
						if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
							*p8Reg |= 0x80;
						}
						SetSignZeroFlag(pReg, *p8Reg);
						SetParity(pReg, *p8Reg);
					break;
					case  0x18: // RR r, 00011RRR
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp = OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						n8Temp = pReg->AFlagReg.RegByte[RegFlag];
						if(*p8Reg & 0x01){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg >>= 1;
						if(n8Temp & FLAG_CARRY) {
							*p8Reg |= 0x80;
						}
						SetSignZeroFlag(pReg, *p8Reg);
						SetParity(pReg, *p8Reg);
					break;
					case  0x20: // SLA r, 00100RRR
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp = OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						if(*p8Reg & 0x80){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg <<= 1;
						SetSignZeroFlag(pReg, *p8Reg);
						SetParity(pReg, *p8Reg);
					break;
					case  0x28: // SRA r, 00101RRR
						pReg->AFlagReg.RegByte[RegFlag] = 0;
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp = OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						n8Temp = *p8Reg;
						if(*p8Reg & 0x01){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg >>= 1;
						*p8Reg |=(n8Temp & 0x80);
						SetSignZeroFlag(pReg, *p8Reg);
						SetParity(pReg, *p8Reg);
					break;
					case  0x38: // SRL r, 00111RRR
						// B 000, C 001, D 010, E 011, H 100, L 101, A 111
						n8Temp = OpCode & 0b00000111;
						if(n8Temp & 1){ // odd
							n8Temp -= 1; 
						}
						else{ // even
							n8Temp += 1;
						}
						p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
						if(*p8Reg & 0x01){
							pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
						}
						else {
							pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
						}
						*p8Reg >>= 1;
						SetSignZeroFlag(pReg, *p8Reg);
						SetParity(pReg, *p8Reg);
					break;
					default:
						if(OpCode >= 0x40){
							// bit 0 & 1 = 40h, bit 2 & 3 = 50h, bit 4 & 5 = 60h, bit 6 & 7 = 70h
							n8Temp2 = (OpCode >> 3) & 7;// bits code now low bits (01bbb110 -> .....bbb)
							if( (OpCode & 7) == 6){ // this is RRR = 6 = BIT b, (HL), 01bbb0110 (xxBBBRRR)
								p8Reg = (uint8_t*)(pMemory + pReg->HLReg.WordReg);
							}
							else{
								n8Temp = OpCode & 7;
								if(n8Temp & 1){ // odd, accounts for C is lower memory than B etc etc
									n8Temp -= 1; 
								}
								else{ // even
									n8Temp += 1;
								}
								p8Reg = (uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp);
							}
							if(OpCode >= 0x40 && OpCode <= 0x7F){ // 01BBBRRR, BIT b, r
								// B 000, C 001, D 010, E 011, H 100, L 101, A 111
								pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;
								pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_N;
								CheckBit(pReg, *p8Reg, n8Temp2);
								break;
							}
							else if(OpCode >= 0x80 && OpCode <= 0xBF){// xxBBBRRR, RES b, r
								n8Temp =1;
								n8Temp <<= n8Temp2;
								*p8Reg &= ~n8Temp;
								break;
							}
							else if(OpCode >= 0xC0 && OpCode <= 0xFF){// xxBBBRRR, SET b, r
								n8Temp =1;
								n8Temp <<= n8Temp2;
								*p8Reg |= n8Temp;
								break;
							}
						} // opcode >= 40h

			}	// inner default switch
		} // outer switch
	}
	pReg->PC++; // done with OpCode
}

	
// LD r, r'
// 01RRRRRR, Hi RRR = Lo RRR
void   LD_R_R(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
	uint8_t nDest8Temp =(OpCode & 0b00111000) >> 3;
	uint8_t nSrc8Temp =OpCode & 0b00000111;
	if(nDest8Temp & 1){ // odd
		nDest8Temp -= 1; 
	}
	else{ // even
		nDest8Temp += 1;
	}
	if(nSrc8Temp & 1){ // odd
		nSrc8Temp -= 1; 
	}
	else{ // even
		nSrc8Temp += 1;
	}
	*(&pReg->BCReg.RegByte[RegC] + nDest8Temp) =  *(&pReg->BCReg.RegByte[RegC] + nSrc8Temp);
	}
}

// LD r, n'
// 00RRR110
void   LD_R_N(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint8_t n8Temp =(OpCode & 0b00111000) >> 3;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		*( &pReg->BCReg.RegByte[RegC] + n8Temp) = (uint8_t) *(pMemory + pReg->PC++);
	}
}

// LD r, (HL)
// 01RRR110
void   LD_R_HL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint8_t n8Temp =(OpCode & 0b00111000) >> 3;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		*( &pReg->BCReg.RegByte[RegC] + n8Temp) = *(pMemory + pReg->HLReg.WordReg);
	}
}

// LD (HL), r
// 01110RRR
void   LD_HL_R(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		*(pMemory + pReg->HLReg.WordReg) = *( &pReg->BCReg.RegByte[RegC] + n8Temp);
	}
}

// LD (HL), n
// 00110110
void   LD_HL_N(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		*(pMemory + pReg->HLReg.WordReg) = (uint8_t) *(pMemory + pReg->PC++);
	}
}

// LD A, (BC)
// 00001010, Ah
void   LD_A_BC(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegA] = *(pMemory + pReg->BCReg.WordReg);
	}
}

// LD A, (DE)
// 00011010
void   LD_A_DE(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegA] = *(pMemory + pReg->DEReg.WordReg);
	}
}

// LD A, (nn)
// 00111010
void   LD_A_NN(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegA] = *(pMemory +( *(pMemory + pReg->PC) | (*(pMemory + pReg->PC+1)<<8)));
		pReg->PC++;
		pReg->PC++;
	}
}

// LD [BC], A, 
// 00000010
void   LD_BC_A(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		*(pMemory + pReg->BCReg.WordReg) = pReg->AFlagReg.RegByte[RegA];
	}
}

// LD [DE], A, 
// 00010010
void   LD_DE_A(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		*(pMemory + pReg->DEReg.WordReg) = pReg->AFlagReg.RegByte[RegA];
	}
}

// LD [nn], A
// 00110010, 32h
void   LD_NN_A(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		*(pMemory +( *(pMemory + pReg->PC) | (*(pMemory + pReg->PC+1)<<8))) = pReg->AFlagReg.RegByte[RegA];
		pReg->PC++;
		pReg->PC++;
	}
}

//LD dd,nn
//00dd0001, 01h,11h,21h,31h, load immediate to Register pair
void   LD_dd_NN(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// Shift by 3 so mult 2 as 2 bytes per reg
		// BC=0,DE=1,HL=2,Stack = 3
		uint8_t n8Temp = (OpCode & 0b00110000) >> 3;
		if(n8Temp == 6){
			n8Temp +=2; // skip Areg & Flags & go to stack
		}
		*((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp)) = *(pMemory + pReg->PC) | ((*(pMemory + pReg->PC+1))<<8);
		pReg->PC +=2;// skip  data
	}
}

// LD HL, [nn]
// 00101010 2Ah
void   LD_HL_MM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t nadd = (uint16_t)*(pMemory + pReg->PC++);
		nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
		pReg->HLReg.RegByte[RegL] = *(pMemory + nadd);
		pReg->HLReg.RegByte[RegH] = *(pMemory + nadd+1);
	}
}

// LD (nn], HL
// 00100010 22h
void   LD_MM_HL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t nadd = (uint16_t)*(pMemory + pReg->PC++);
		nadd |= (uint16_t)((*(pMemory + pReg->PC++))<< 8);
		*(pMemory + nadd) = pReg->HLReg.RegByte[RegL];
		*(pMemory + nadd+1) = pReg->HLReg.RegByte[RegH];
	}
}


// LD SP, HL
// 11111001 F9h
void   LD_SP_HL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->StackPtr.WordReg = pReg->HLReg.WordReg;
	}
}

// PUSH qq
// 11qq0101
void   PUSH_QQ(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
	// dd = BC=00,DE=01, HL=10, AF = 11
	// First Dec SP then store Hi byte first then dec SP and store Lo byte, Store A reg dec and store Flags
		pReg->StackPtr.WordReg--; // Dec SP
		//uint8_t nReg = ( (OpCode & 0b00110000) >> 4);
		switch (( (OpCode & 0b00110000) >> 4)){
			case 0:
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->BCReg.RegByte[RegB];
				pReg->StackPtr.WordReg--;
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->BCReg.RegByte[RegC];
				break;
			case 1:
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->DEReg.RegByte[RegD];
				pReg->StackPtr.WordReg--;
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->DEReg.RegByte[RegE];
				break;
			case 2:
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->HLReg.RegByte[RegH];
				pReg->StackPtr.WordReg--;
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->HLReg.RegByte[RegL];
				break;
			case 3:
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->AFlagReg.RegByte[RegA];
				pReg->StackPtr.WordReg--;
				*((uint8_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->AFlagReg.RegByte[RegFlag];
				break;	
		}
	}
}

// POP qq
// 11qq0001
void   POP_QQ(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
	// dd = BC=00,DE=01, HL=10, AF = 11
	// copy SP data to Lo byte, Inc SP then copy SP data to Hi Byte then INC , Flags First the A reg
		switch (( (OpCode & 0b00110000) >> 4)){
			case 0:
			pReg->BCReg.RegByte[RegC] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			pReg->BCReg.RegByte[RegB] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			break;
			case 1:
			pReg->DEReg.RegByte[RegE] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			pReg->DEReg.RegByte[RegD] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			break;
			case 2:
			pReg->HLReg.RegByte[RegL] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			pReg->HLReg.RegByte[RegH] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			break;
			case 3:
			pReg->AFlagReg.RegByte[RegFlag] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			pReg->AFlagReg.RegByte[RegA] = *((uint8_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg++;
			break;
		}
	}
}

// EX DE, HL
// 11101011, EBh
void   XCHG_DE_HL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t TempWord = pReg->HLReg.WordReg;
		pReg->HLReg.WordReg = pReg->DEReg.WordReg;
		pReg->DEReg.WordReg = TempWord;
	}
}

// EX AF, 'AF
// 00001000 , 08h
void   XCHG_AF_ALTAF(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
	
		uint16_t TempWord = pReg->AltAFlagReg.WordReg;
		pReg->AltAFlagReg.WordReg = pReg->AFlagReg.WordReg;
		pReg->AFlagReg.WordReg = TempWord;
	}
}

// EXX  BC, DE & HL exchanged with Alt Regs
// 11011001 , D9h
void   XCHG_WREGS_ALTWREGS(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint16_t TempWord = pReg->AltBCReg.WordReg;
		pReg->AltBCReg.WordReg = pReg->BCReg.WordReg;
		pReg->BCReg.WordReg = TempWord;
		TempWord = pReg->AltDEReg.WordReg;
		pReg->AltDEReg.WordReg = pReg->DEReg.WordReg;
		pReg->DEReg.WordReg = TempWord;
		TempWord = pReg->AltHLReg.WordReg;
		pReg->AltHLReg.WordReg = pReg->HLReg.WordReg;
		pReg->HLReg.WordReg = TempWord;
	}
}

// EX (SP), HL
// 11100011 , E3h
void   XCHG_SPADD_HL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 

		uint16_t TempByte1 = pReg->HLReg.RegByte[RegL];
		uint16_t TempWord2 = pReg->HLReg.RegByte[RegH];
		pReg->HLReg.RegByte[RegL] = *(pMemory + pReg->StackPtr.WordReg);// lo byte
		pReg->HLReg.RegByte[RegH] = *(pMemory + pReg->StackPtr.WordReg +1); // hi byte
		*(pMemory + pReg->StackPtr.WordReg) = TempByte1; // low byte
		*(pMemory + pReg->StackPtr.WordReg+1) = TempWord2; // hi byte
	}
}



// ADD A, r
// 10000rrr 
void   ADD_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get other reg
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		uint8_t regx = *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], regx, NO_CARRY);
		}
}

// ADD A, n
// 11000110, C6 
void   ADD_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get nn
		Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->PC++),NO_CARRY );
	}
}

//ADD A,(HL)
// 10000110, 86h 
void   ADD_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get nn
		Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->HLReg.WordReg), NO_CARRY);
	}
}

// ADC A, r
// 10001rrr 
void   ADC_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get other reg
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		uint8_t regx = *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], regx, INCLUDE_CARRY);
	}
}

// ADC A, n
// 11001110, CE 
void   ADC_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->PC++),INCLUDE_CARRY);
	}
}

//ADC A,(HL)
// 10001110, 8E 
void   ADC_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		Add_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->HLReg.WordReg), INCLUDE_CARRY);
	}
}

// SUB A, r
// 10010rrr 
void   SUB_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get other reg
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		uint8_t regx = *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], regx, NO_CARRY );
		}
}

// SUB A, n (A-n)
// 11010110, D6 
void   SUB_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get nn
		Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->PC++),NO_CARRY );
	}
}

//SUB A,(HL)
// 10010110, 96 
void   SUB_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get nn
		Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->HLReg.WordReg), NO_CARRY);
	}
}

// SBC A, r
// 10011rrr 
void   SBC_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// get other reg
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		uint8_t regx = *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], regx, INCLUDE_CARRY);
	}
}

// SBC A, n
// 11011110, DE 
void   SBC_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->PC++),INCLUDE_CARRY);
	}
}

//SBC A,(HL)
// 10011110, 9E 
void   SBC_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		Sub_8Bit(pReg, &pReg->AFlagReg.RegByte[RegA], *(pMemory + pReg->HLReg.WordReg), INCLUDE_CARRY);
	}
}

////////////////////////
// AND A, r
// 10100rrr 
void   AND_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;// set H
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		pReg->AFlagReg.RegByte[RegA] &= *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

// AND A, n
// 11100110, E6 
void   AND_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;// set H
		pReg->AFlagReg.RegByte[RegA] &= *(pMemory + pReg->PC++);
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

//AND A,(HL)
// 10100110, A6 
void   AND_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY;// set H
		pReg->AFlagReg.RegByte[RegA] &= *(pMemory + pReg->HLReg.WordReg);
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

// OR A, r
// 10110rrr 
void   OR_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		pReg->AFlagReg.RegByte[RegA] |= *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

// OR A, n
// 11110110, F6 
void   OR_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegA] |= *(pMemory + pReg->PC++);
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

//OR A,(HL)
// 10110110, B6 
void   OR_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegA] |= *(pMemory + pReg->HLReg.WordReg);
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

// XOR r
// 10101rrr 
void   XOR_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		uint8_t n8Temp = OpCode & 0b00000111;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		pReg->AFlagReg.RegByte[RegA] ^= *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

// XOR A, n
// 11101110, EE 
void   XOR_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegA] ^= *(pMemory + pReg->PC++);
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

//XOR A,(HL)
// 10101110, AE 
void   XOR_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] = 0;
		pReg->AFlagReg.RegByte[RegA] ^= *(pMemory + pReg->HLReg.WordReg);
		SetParity(pReg, pReg->AFlagReg.RegByte[RegA]);
		SetSignZeroFlag(pReg, pReg->AFlagReg.RegByte[RegA]);
	}
}

// CP A, r
// 10111rrr 
void   CP_rr(unsigned int* padd, uint8_t OpCode, struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
			uint8_t n8Temp = OpCode & 0b00000111;
			if(n8Temp & 1){ // odd
				n8Temp -= 1; 
			}
			else{ // even
				n8Temp += 1;
			}
			CMP_ARegHandler(pReg, *((uint8_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp)));
	}
}

// CP A, n
// 11111110, FE 
void   CP_n(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
			CMP_ARegHandler(pReg, *(pMemory + pReg->PC++));
	}
}

//CP A,(HL)
// 10111110, BE 
void   CP_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		CMP_ARegHandler(pReg, *(pMemory + pReg->HLReg.WordReg));	
	}
}

// INC r
// 00RRR100
void   INC_R(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint8_t n8Temp = (OpCode & 0b00111000) >> 3;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		uint8_t* pregbyteadd = &pReg->BCReg.RegByte[RegC] + n8Temp;
		Inc8BitVar(pReg, pregbyteadd);
	}
}

// INC (HL)
// 00110100,34h
void   INC_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		Inc8BitVar(pReg, (pMemory + pReg->HLReg.WordReg));
	}
}


// DEC r
// 00RRR101
void   DEC_R(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		uint8_t n8Temp = (OpCode & 0b00111000) >> 3;
		if(n8Temp & 1){ // odd
			n8Temp -= 1; 
		}
		else{ // even
			n8Temp += 1;
		}
		uint8_t* pregbyteadd = &pReg->BCReg.RegByte[RegC] + n8Temp;
		Dec8BitVar(pReg, pregbyteadd);
	}
}

// DEC (HL)
// 00110101,35h
void   DEC_MEM(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		Dec8BitVar(pReg, (pMemory + pReg->HLReg.WordReg));
	}
}

// DAA 
// 00100111,27h
void   DAA(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		DAAMngr(pReg);
	}
}

//CPL
// 00101111,2Fh
void   CPL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegA] = ~pReg->AFlagReg.RegByte[RegA];
		pReg->AFlagReg.RegByte[RegFlag] |= 	(FLAG_N | FLAG_HCARRY);
	}
}

//CCF
// 00111111,3Fh
void   CCF(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_N;
		pReg->AFlagReg.RegByte[RegFlag] ^= 	FLAG_CARRY; // flip carry 
	}
}

//SCF
// 00110111,37h
void   SCF(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] &= 	(~FLAG_N | ~FLAG_HCARRY);
		pReg->AFlagReg.RegByte[RegFlag] |= 	FLAG_CARRY;
	}
}

//NOP
// 00000000,00h
void   NOP(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		asm("nop");
	}
}

//HLT
// 01110110,76h
void   HLT(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// will return to MMBasic program at this stage
		pReg->ProgCtrlFlags |= FLAG_EXIT;
	}
}

//DI
// 11110011,F3h
void   DI(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->IFF &= 0b11111110;
	}
}

//EI
// 11111011,FBh
void   EI(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->IFF |= 1; 
	}
}

// ADD HL,ss
// 00SS1001
void   ADD_HL_SS(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// Shift by 3 so mult 2 as 2 bytes per reg
		uint8_t n8Temp = (OpCode & 0b00110000) >> 3;
		if(n8Temp == 6){
			n8Temp +=2; // skip Areg & Flags
		}
		uint16_t RegVal = *((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp));
		if( ((pReg->HLReg.WordReg & 0x07FF) + (RegVal & 0x07FF)) > 0x7FF){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_HCARRY; // set
		}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= ~FLAG_HCARRY; // reset
		}
		if(  ((uint32_t) pReg->HLReg.WordReg + (uint32_t) RegVal) > 0xFFFF ){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
		}
		pReg->HLReg.WordReg += RegVal;
		pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_N;
	}
}

// INC,ss
// 00SS0011
void   INC_SS(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	//BC=0,DE=1,HL=2, Stack=3
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// Shift by 3 so mult 2 as 2 bytes per reg
		uint8_t n8Temp = (OpCode & 0b00110000) >> 3;
		if(n8Temp == 6){
			n8Temp +=2; // skip Areg & Flags
		}
		*((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp)) += 1;
	}
	
}

// DEC,ss
// 00SS1011
void   DEC_SS(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		// Shift by 3 so mult 2 as 2 bytes per reg
		uint8_t n8Temp = (OpCode & 0b00110000) >> 3;
		if(n8Temp == 6){
			n8Temp +=2; // skip Areg & Flags
		}
		*((uint16_t*)(&pReg->BCReg.RegByte[RegC] + n8Temp)) -= 1;
	}
	
}

// RLCA
// 00000111, 07h
void   RLCA(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] &= FLAG_H_AND_N_RESET;
		if(pReg->AFlagReg.RegByte[RegA] & 0x80){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
		}
		pReg->AFlagReg.RegByte[RegA] <<= 1;
		pReg->AFlagReg.RegByte[RegA] |= pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY;
	}
	
}

// RLA
// 00010111, 17h
void   RLA(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] &= FLAG_H_AND_N_RESET;
		uint8_t ncarrytemp = pReg->AFlagReg.RegByte[RegFlag];
		if(pReg->AFlagReg.RegByte[RegA] & 0x80){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
		}
		pReg->AFlagReg.RegByte[RegA] <<= 1;
		pReg->AFlagReg.RegByte[RegA] |= (ncarrytemp & FLAG_CARRY);
	}
}

// RRCA
// 00001111, 0Fh
void   RRCA(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] &= FLAG_H_AND_N_RESET;
		if(pReg->AFlagReg.RegByte[RegA] & 0x01){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
		}
		pReg->AFlagReg.RegByte[RegA] >>= 1;
		if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
			pReg->AFlagReg.RegByte[RegA] |= 0x80;
		}
	}
}

// RRA
// 00011111, 1Fh
void   RRA(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->AFlagReg.RegByte[RegFlag] &= FLAG_H_AND_N_RESET;
		uint8_t ncarrytemp = pReg->AFlagReg.RegByte[RegFlag];
		if(pReg->AFlagReg.RegByte[RegA] & 0x01){
			pReg->AFlagReg.RegByte[RegFlag] |= FLAG_CARRY;
		}
		else {
			pReg->AFlagReg.RegByte[RegFlag] &= 	~FLAG_CARRY;
		}
		pReg->AFlagReg.RegByte[RegA] >>= 1;
		if(ncarrytemp & FLAG_CARRY){
			pReg->AFlagReg.RegByte[RegA] |= 0x80;
		}
	}
}

// JP nn
// C3
void   JP_NN(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		uint16_t TempWord;
		TempWord = *(pMemory + pReg->PC++);
		TempWord |= (*(pMemory + pReg->PC) << 8);
		pReg->PC = TempWord;
	}
}

// JP cc, nn
// 11CCC010
void   JP_CC(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		uint8_t bJump =0;
		uint16_t	TempWord;
		switch( (OpCode & 0b00111000) >> 3){
			case 0: // Non-Zero
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO)==0){
					bJump = 1;
				}
			break;
			case 1: // Zero
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO){
					bJump = 1;
				}
			break;
			case 2: // No Carry
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY)==0){
					bJump = 1;
				}
			break;
			case 3: // Carry
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
					bJump = 1;
				}
			break;
			case 4: // Parity Odd
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_PV)==0){
					bJump = 1;
				}
			break;
			case 5: // Parity Even
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_PV){
					bJump = 1;
				}
			break;
			case 6: // Sign Positive
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_SIGN)==0){
					bJump = 1;
				}
			break;
			case 7: // Sign Negative
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_SIGN){
					bJump = 1;
				}
			break;
		}  // switch
		if(bJump){
			TempWord = *(pMemory + pReg->PC++);
			TempWord |= (*(pMemory + pReg->PC) << 8);
			pReg->PC = TempWord;
		}
		else{
			pReg->PC +=2;
		}
	} // else
}
		
// JR e
// 18n (2 bytes)
void   JR_E(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{ 
		pReg->PC--;// reset to start of instruction before offsetting
		if((*(pMemory + pReg->PC+1)) & 0x80){
			pReg->PC -= 0x100 - *(pMemory + pReg->PC+1);
		}
		else {
			pReg->PC += *(pMemory + pReg->PC+1);	
		}
	}
}

// JR C e
// 38
void   JRC_E(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
			pReg->PC--;// reset to start of instruction before offsetting
			if((*(pMemory + pReg->PC+1)) & 0x80){
				pReg->PC -= 0x100 - *(pMemory + pReg->PC+1);
			}
			else {
				pReg->PC += *(pMemory + pReg->PC+1);	
			}
		}
		else{
			pReg->PC++;
		}
	}
}

// JR NC e
// 30
void   JRNC_E(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		if( (pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY) == 0){
			pReg->PC--;// reset to start of instruction before offsetting
			if((*(pMemory + pReg->PC+1)) & 0x80){
				pReg->PC -= 0x100 - *(pMemory + pReg->PC+1);
			}
			else {
				pReg->PC += *(pMemory + pReg->PC+1);	
			}
		}
		else{
			pReg->PC++;
		}
	}
}

// JR Z e
// 28
void   JRZ_E(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO){
			pReg->PC--;// reset to start of instruction before offsetting
			if((*(pMemory + pReg->PC+1)) & 0x80){
				pReg->PC -= 0x100 - *(pMemory + pReg->PC+1);
			}
			else {
				pReg->PC += *(pMemory + pReg->PC+1);	
			}
		}
		else{
			pReg->PC++;
		}
	}
}

// JR NZ e
// 20
void   JRNZ_E(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		if( (pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO) == 0){
			pReg->PC--;// reset to start of instruction before offsetting
			if((*(pMemory + pReg->PC+1)) & 0x80){
				pReg->PC -= 0x100 - *(pMemory + pReg->PC+1);
			}
			else {
				pReg->PC += *(pMemory + pReg->PC+1);	
			}
		}
		else{
			pReg->PC++;
		}
	}
}

// JP (HL)
// E9
void   JP_HL(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		pReg->PC = pReg->HLReg.WordReg;
	}
}

// DJ NZ e
// 10
void   DJNZ_E(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		if( --pReg->BCReg.RegByte[RegB] != 0){
			pReg->PC--;// reset to start of instruction before offsetting
			if((*(pMemory + pReg->PC+1)) & 0x80){
				pReg->PC -= 0x100 - *(pMemory + pReg->PC+1);
			}
			else {
				pReg->PC += *(pMemory + pReg->PC+1);	
			}
		}
		else{
			pReg->PC++;// skip 2nd byte of instruction
		}
	}
	
}

// Call nn
// CD
void   CALL_NN(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		// save current PC on stack & load PC with address
		pReg->StackPtr.WordReg -=2;
		*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC+2; // when pops is next instruction
		uint8_t TempByte1 = *(pMemory + pReg->PC++);
		uint8_t	TempByte2 = *(pMemory + pReg->PC);
		pReg->PC = TempByte1;
		pReg->PC |= TempByte2 << 8;
	}
}

// CALL cc, nn
// 11CCC010
void   CALL_CC_NN(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		uint8_t bCall =0;
		switch( (OpCode & 0b00111000) >> 3){
			case 0: // Non-Zero
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO)==0){
					bCall = 1;
				}
			break;
			case 1: // Zero
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO){
					bCall = 1;
				}
			break;
			case 2: // No Carry
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY)==0){
					bCall = 1;
				}
			break;
			case 3: // Carry
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
					bCall = 1;
				}
			break;
			case 4: // Parity Odd
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_PV)==0){
					bCall = 1;
				}
			break;
			case 5: // Parity Even
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_PV){
					bCall = 1;
				}
			break;
			case 6: // Sign Positive
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_SIGN)==0){
					bCall = 1;
				}
			break;
			case 7: // Sign Negative
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_SIGN){
					bCall = 1;
				}
			break;
		}  // switch
		if(bCall){
			// save current PC on stack & load PC with address
			pReg->StackPtr.WordReg -=2;
			*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC+2; // when pops is next instruction
			uint8_t TempByte1 = *(pMemory + pReg->PC++);
			uint8_t	TempByte2 = *(pMemory + pReg->PC);
			pReg->PC = TempByte1;
			pReg->PC |= TempByte2 << 8;
		}
		else{
			pReg->PC +=2;
		}
	} // else
}

// RET
// C9
void   RET(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		// POP off Stack and place in PC
		pReg->PC = *((uint16_t*)(pMemory + pReg->StackPtr.WordReg));
		pReg->StackPtr.WordReg += 2;
	}
}

// RET cc, nn
// 11CCC000
void   RET_CC(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		uint8_t bRet =0;
		switch( (OpCode & 0b00111000) >> 3){
			case 0: // Non-Zero
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO)==0){
					bRet = 1;
				}
			break;
			case 1: // Zero
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_ZERO){
					bRet = 1;
				}
			break;
			case 2: // No Carry
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY)==0){
					bRet = 1;
				}
			break;
			case 3: // Carry
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_CARRY){
					bRet = 1;
				}
			break;
			case 4: // Parity Odd
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_PV)==0){
					bRet = 1;
				}
			break;
			case 5: // Parity Even
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_PV){
					bRet = 1;
				}
			break;
			case 6: // Sign Positive
				if((pReg->AFlagReg.RegByte[RegFlag] & FLAG_SIGN)==0){
					bRet = 1;
				}
			break;
			case 7: // Sign Negative
				if(pReg->AFlagReg.RegByte[RegFlag] & FLAG_SIGN){
					bRet = 1;
				}
			break;
		}  // switch
		if(bRet){
			// POP off Stack and place in PC
			pReg->PC = *((uint16_t*)(pMemory + pReg->StackPtr.WordReg));
			pReg->StackPtr.WordReg += 2;
		}
	} // else
}

//RST p
void   RST_P(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		// pushes the contents of the program counter onto the stack to provide a return address and then Jumps to 
		// the code set in bits 3,4&5
		pReg->StackPtr.WordReg -=2;
		*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC;
		// now setPC to correct RST location
		pReg->PC = (uint16_t )(OpCode & 0x38); // leave bits 3,4 &5
	}
}

//IN A (n), n is the CMM2 port number
//DBh
void   IN_A_N(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		pReg->AFlagReg.RegByte[RegA] = (uint8_t) PinRead((uint32_t) OpCode);
	}
}

//OUT (n), A  n is the CMM2 port number
//D3h
void   OUT_N_A(unsigned int* padd, uint8_t OpCode,struct Registers* pReg, uint8_t* pMemory){
	if(pReg->nFuncInit){ // initialise the address
		GetRetAdd(padd);
	}
	else{  
		// This program outs the data in accumulator to the selected pin defined in the operand, pins as specified in the
		// CMM2 manual (1 to 40, excluding power and grounds).  Its operators responsibility to ensure the pin is set to
		// DOUT in the main MMBasic program prior to calling this.
		ExtSet((int)pReg->AFlagReg.RegByte[RegA],(int) OpCode);
	}
}


void HandleCode(struct Registers* pReg, uint8_t* pMemory, uint8_t* pFuncMem) //__attribute__((aligned (32)))
{
	// PC pointing to instruction
	// will exit with PC pointing to next instruction
	// pMemory points to the 64k of memory so can be source code or variable memory pointed to by stack
	// Note stack stores data in little endian, i.e, StackPtr->LSB, StackPtr+1->MSB
	uint8_t OpCode=0;
	uint64_t nRoutineChecksTimer64;
	uint64_t nIntChecks64;
	uint32_t nMsecCount=0;
	unsigned int* padd=0;
	uint8_t nJ;
	uint8_t nK;
	uint8_t nCnt;
	char nchar;
	
	// create Pointer to array of pointers to functions
	void (*(*pFunArr)[256])(unsigned int* , uint8_t , struct Registers* , uint8_t* ); // ptr to array of functions
	pFunArr = (void (*(*)[256])(unsigned int* , uint8_t , struct Registers* , uint8_t* )) pFuncMem;// convert MMBasic memory to function array
	if ( !(pReg->ProgCtrlFlags & FLAG_INIT)){ 
		pReg->cMDSK[0] = 'M'; // set MMBasic call back
		pReg->cMDSK[1] = 'M';
		pReg->cMDSK[2] = 'C';
		pReg->cMDSK[3] = 'B';
		pReg->cMDSK[4] = '\0'; //  terminate with 2 null chars
		pReg->cMDSK[5] = '\0';
		pReg->nError = 0;
		#ifdef MSEC_COUNT_INCLUDED
		*((uint8_t*)(pMemory + COUNT_FLAG_ADDRESS))	= 0; // turn off getting now time
		nMsecCount = 0;
		#endif
		#ifdef COUNT_INSTRUCT_WHILE_TIMING
		// TEST ONLY ---------------------		
		pReg->nUserVar_28_8 =0; // test only	
		// ------------------------------	
		#endif
		// will get Function addresses and store in array
		// create function pointers
		void (*fun_ptr)(unsigned int* , uint8_t ,struct Registers* , uint8_t* );
		pReg->nFuncInit = 1; // *** initialise address
		// Multi Function OpCodes
		// *** 1101 1101 (DD) ********************
		DDIXHandler(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xDD] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_20);
		#endif
		// *** EDHandler, 11101101, ED ********************
		EDHandler(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0b11101101] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_20);
		#endif
		// ***  1111 1101 (FD) ********************
		FDIYHandler(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xFD] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_20);
		#endif
		// ***  11001011, CB ********************
		CBHandler(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xCB] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_20);
		#endif
	
		//  8 BIT LOAD GROUP Opcodes  ---------------------------------------------
		// *** LD r, r , // 01RRRRRR ********************
		LD_R_R(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				for(nK=0;nK<=7;nK++){
					if(nK != 6) {
						nCnt = nK + ((nJ<<3) | 0b01000000);	// 0x40 start of LD r,r
						(*pFunArr)[nCnt] = fun_ptr;
					}
				}
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD r, n, 00RRR110 ********************
		LD_R_N(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
			nCnt = ((nJ<<3) | 0b00000110); //0x06 start of LD r,n	


			(*pFunArr)[nCnt] = fun_ptr;
	
			}
		} 

		// *** LD r, (HL), 01RRR110 ********************
		LD_R_HL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
			nCnt = ((nJ<<3) | 0b01000110); //0x46 start of LD r,n	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		}
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// *** LD [HL], r 01110RRR ********************
		LD_HL_R(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
			nCnt = (nJ | 0b01110000); //0x70 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD (HL), n 00110110 ********************
		LD_HL_N(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x36] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD A, (BC), 00001010 , Ah ********************
		LD_A_BC(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xA] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD A, (DE), 00011010 ********************
		LD_A_DE(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x1A] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD_A_NN, 00111010 ********************
		LD_A_NN(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x3A] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD_BC_A, 00000010 ********************
		LD_BC_A(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x02] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD_DE_A, 00010010 ********************
		LD_DE_A(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x12] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD_NN_A, 00110010 ********************
		LD_NN_A(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x32] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		
		// 16 BIT LOAD GROUP Opcodes --------------------------------------------
		// *** LD dd,n,n, 00dd0001 ********************
		LD_dd_NN(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=3;nJ++){
			nCnt = ((nJ<<4) | 0b0000001); //0x01 
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd,JUMP_18 );
		#endif
		
		// *** LD_HL_MM, 00101010, 2Ah ********************
		LD_HL_MM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x2A] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD_MM_HL, 00100010, 22h ********************
		LD_MM_HL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x22] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** LD_SP_HL, 11111001, F9h ********************
		LD_SP_HL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xF9] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** PUSH_QQ, 11qq0101,  ********************
		PUSH_QQ(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=3;nJ++){
			nCnt = ((nJ<<4) | 0b11000101); //0xC5 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** POP_QQ, 11qq0001,  ********************
		POP_QQ(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=3;nJ++){
			nCnt = ((nJ<<4) | 0b11000001); //0xC1 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
			
		// EXCHANGE, BLOCK TRANSFER AND SEARCH GROUP -------------------------------------
		// *** XCHG_DE_HL, 11101011, EBh ********************
		XCHG_DE_HL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xEB] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** XCHG_AF_ALTAF, 00001000 , 08h ********************
		XCHG_AF_ALTAF(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x08] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** XCHG_WREGS_ALTWREGS, 11011001 , D9h ********************
		XCHG_WREGS_ALTWREGS(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xD9] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** XCHG_SPADD_HL, 11100011 , E3h ********************
		XCHG_SPADD_HL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xE3] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		
		// 8 BIT ARIlTHMATIC AND LOGICAL GROUP Opcodes --------------------------------------
		// *** ADD_rr, 10000rrr,  ********************
		ADD_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
			nCnt = (nJ | 0x80); //0x80 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
				
		// *** ADD_n, 11000110, C6h ********************
		ADD_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xC6] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** ADD_MEM, 10000110, 86h ********************
		ADD_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x86] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** ADC_rr, 10001rrr,  ********************
		ADC_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0x88); //0x88 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** ADC_n, 11001110, CE  ********************
		ADC_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xCE] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** ADC_MEM, 10001110, 8E  ********************
		ADC_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x8E] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SUB_rr, 10010rrr,  ********************
		SUB_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0x90); //0x90 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SUB_n, 11010110, D6  ********************
		SUB_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xD6] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SUB_MEM, 10010110, 96   ********************
		SUB_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x96] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SBC_rr, 10011rrr,  ********************
		SBC_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0x98); //0x98 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SBC_n, 11011110, DE    ********************
		SBC_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xDE] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SBC_MEM, 10011110, 9E     ********************
		SBC_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x9E] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** AND_rr, 10100rrr     ********************
		AND_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0xA0); //0xA0 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		
		// *** AND_n, 11100110, E6     ********************
		AND_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xE6] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** AND_MEM, 0100110, A6      ********************
		AND_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xA6] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** OR_rr, 10100rrr     ********************
		OR_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0xB0); //0xB0 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// *** OR_n, 11110110, F6     ********************
		OR_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xF6] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** OR_MEM, 10110110, B6      ********************
		OR_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xB6] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** XOR_rr, 10101rrr     ********************
		XOR_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0xA8); //0xA8 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** XOR_n, 11101110, EE     ********************
		XOR_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xEE] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** XOR_MEM, 10101110, AE      ********************
		XOR_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xAE] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** CP_rr, 10101rrr     ********************
		CP_rr(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = (nJ | 0xB8); //0xB8 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// *** CP_n, 11111110, FE     ********************
		CP_n(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xFE] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** CP_MEM, 10111110, BE      ********************
		CP_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xBE] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** INC_R, 00RRR100     ********************
		INC_R(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = ( (nJ<<3) | 0x04); //0x04 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

							
		// *** INC_MEM, 00110100,34h     ********************
		INC_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x34] = fun_ptr;
		// *** DEC_R, 00RRR101     ********************
		DEC_R(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			if(nJ != 6){
				nCnt = ((nJ<<3) | 0x05); //0x05 	
				(*pFunArr)[nCnt] = fun_ptr;
			}
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** DEC_MEM, 00110101,35h     ********************
		DEC_MEM(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x35] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		
		// -GENERAL PURPOSE ARITHMATIC AND CPU CONTROL GROUPS ------------------------------------------------
		// *** DAA, 00100111,27h     ********************
		DAA(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x27] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** CPL, 00101111,2Fh    ********************
		CPL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x2F] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** CCF, 00111111,3Fh    ********************
		CCF(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x3F] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** SCF, 00110111,37h    ********************
		SCF(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x37] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** NOP, 00000000,00h    ********************
		NOP(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** HLT, 01110110,76h    ********************
		HLT(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x76] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** DI, 11110011,F3h    ********************
		DI(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xF3] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** EI, 11111011,FBh    ********************
		EI(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xFB] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// -GENERAL PURPOSE 16 bit ARITHMATIC GROUP ------------------------------------------------
		// *** ADD_HL_SS, 00SS1001,  ********************
		ADD_HL_SS(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=3;nJ++){
			nCnt = ((nJ<<4) | 0x09); //0x09 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** INC_SS, 00SS0011,  ********************
		INC_SS(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=3;nJ++){
			nCnt = ((nJ<<4) | 0x03); //0x03 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** DEC_SS, 00SS0011,  ********************
		DEC_SS(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=3;nJ++){
			nCnt = ((nJ<<4) | 0x0B); //0x0B 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		
		// -Rotate and Shift Group  ------------------------------------------------
		// *** RLCA, 00000111, 07h    ********************
		RLCA(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x07] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** RLA, 00010111, 17h   ********************
		RLA(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x17] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** RRCA, 00001111, 0Fh   ********************
		RRCA(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x0F] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** RRA, 00011111, 1Fh   ********************
		RRA(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x1F] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// -Jump Group Group  ------------------------------------------------
		// *** JP_NN, C3h   ********************
		JP_NN(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xC3] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** JP_CC, 11CCC010,  ********************
		JP_CC(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			nCnt = ((nJ<<3) | 0xC2); //0xC2 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
							
		// *** JR_E, 18h   ********************
		JR_E(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x18] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** JRC_E, 38h   ********************
		JRC_E(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x38] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** JRNC_E, 30h   ********************
		JRNC_E(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x30] = fun_ptr;
		// *** JRZ_E, 28h   ********************
		JRZ_E(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x28] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** JRNZ_E, 20h   ********************
		JRNZ_E(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x20] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** JP_HL, E9h   ********************
		JP_HL(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xE9] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** DJNZ_E, 10h   ********************
		DJNZ_E(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0x10] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// Call and Return Group  ------------------------------------------------
		// *** CALL_NN, CDh   ********************
		CALL_NN(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xCD] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** CALL_CC_NN, 11CCC100,  ********************
		CALL_CC_NN(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			nCnt = ((nJ<<3) | 0xC4); //0xC4 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** RET, C9h   ********************
		RET(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xC9] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** RST_P, 11TTT111,  ********************
		RST_P(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			nCnt = ((nJ<<3) | 0xC7); //0xC7 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** RET_CC, 11CCC000, conditional returns ********************
		RET_CC(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		for(nJ=0;nJ<=7;nJ++){
			nCnt = ((nJ<<3) | 0xC0); //0xC0 	
			(*pFunArr)[nCnt] = fun_ptr;
		} 
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif

		// -IN OUT Group  ------------------------------------------------
		// *** IN_A_N, DBh   ********************
		IN_A_N(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xDB] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		// *** OUT_N_A, D3h   ********************
		OUT_N_A(padd, OpCode,pReg, pMemory);
		*padd -= 32; // *** 32 = number of bytes from calling GetRetAdd() in function entry 
		fun_ptr = (void (*)(unsigned int* , uint8_t ,struct Registers* , uint8_t* )) *padd; // *** set pointer
		(*pFunArr)[0xD3] = fun_ptr;
		#ifdef OVERWRITE_GET_ADDRESS
			ByPassGetAddress(padd, JUMP_18);
		#endif
		
		pReg->nFuncInit = 0; // *** End of getting addresses
		pReg->ProgCtrlFlags  |= FLAG_INIT; // ensure we only do once
	} // end Prog Init
	else{

		nRoutineChecksTimer64 = uSecTimer();
		nIntChecks64 = nRoutineChecksTimer64;
		while(1){
			OpCode = *(pMemory + pReg->PC); 
			if(pReg->LogOn){// will only log in Step and Run 
				pReg->MMBParameter = MMB_TRACE_REG_STATUS;
				RunBasicSub(pReg->cMDSK);
			}
			pReg->PC++; // skip instruction
			(*pFunArr)[OpCode](padd, OpCode,pReg, pMemory); // run assembly instruction	
			if( pReg->ProgCtrlFlags & FLAG_EXIT) {
				return; // exit
			}
			if ( pReg->ProgCtrlFlags & FLAG_RUN_TO){
				if(pReg->nRunToAddress == pReg->PC)
					return;// at stop address
			}
			else if ( pReg->ProgCtrlFlags & FLAG_STEP){
				if(--pReg->nDebugstep == 0)
					return; // if steps completed
			}
			if( (uSecTimer() - nRoutineChecksTimer64) > 999) { // do checks every millisecond
				RoutineChecks();//for CMM2 internals
				// check for CTRL Q here as well
				nchar = (char) getConsole();// getconsole return -1 (ff) if no char else will return char off MMBASIC buffer
				if(nchar != 0xFF){ // we have a char, save it cos we are only checking CTRL Q
					if(nchar != 0xA){
						pReg->KbBuffer[pReg->KBTail++] = nchar;
						pReg->LastKeyPress =nchar;
						if(pReg->KBTail == KB_BUFF_SIZE){
							pReg->KBTail = 0;
						}
					}// not LF, dont process line feed
					else{
						pReg->LastKeyPress =0;
					}
				} // have char in MMBasic buffer
				if(pReg->LastKeyPress == 17) // check if CTRL Q
					return; // exit
				nRoutineChecksTimer64 = uSecTimer(); 
				nMsecCount++;
			} // every millisecond
			#ifdef MSEC_COUNT_INCLUDED
			// get elapsed timers if COUNT_FLAG_ADDRESS location set
			if( *((uint8_t*) (pMemory + COUNT_FLAG_ADDRESS))){	
				if( *((uint8_t*) (pMemory + COUNT_FLAG_ADDRESS)) == COUNT_FLAG_NOW){
					*((uint32_t*)(pMemory + COUNTER32_ADDRESS)) = nMsecCount;

				}
				else if( *((uint8_t*) (pMemory + COUNT_FLAG_ADDRESS)) == COUNT_FLAG_ELAPSE){	
					*((uint32_t*)(pMemory + COUNTER32_ADDRESS)) = nMsecCount - *((uint32_t*)(pMemory + COUNTER32_ADDRESS));
				}
				*((uint8_t*) (pMemory + COUNT_FLAG_ADDRESS)) = 0; // only do once
			}
			#endif
			
			if(pReg->nIntTime != 0xFFFF){ // only check ints if requested
					if( (pReg->nIntTime == 0) || ((uSecTimer() - nIntChecks64) >= pReg->nIntTime) ) {
						// NMI
						if(pReg->nNMIPin){
							if(PinRead(pReg->nNMIPin)==0){
								pReg->StackPtr.WordReg -=2;
								*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC;
								// now setPC to correct RST location
								pReg->PC = 0x66;
							} // NMI
						} // NMI Int pin set	
						if(pReg->IFF & 1){ // are ints enabled
							switch((pReg->IFF & 0b11000000)){
								case 0:
									Mode_0_INTHandler(pReg, pMemory);
								break;
								case 1:
									pReg->StackPtr.WordReg -=2;
									*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC;
									// now setPC to correct RST location
									pReg->PC = 0x38;
								break;
								case 2:
									pReg->StackPtr.WordReg -=2;
									*((uint16_t*)(pMemory + pReg->StackPtr.WordReg)) = pReg->PC;
									pReg->PC = (pReg->IReg<<4) | pReg->IntMode_2_Offset; // PC now at correct restart table address
									break;
							}
						} // only do if ints enabled
						nIntChecks64 = uSecTimer();
					}
			} // if ints active from MMBasic program
		}// while for ever
	} // else not init

}


// MMBASIC Entry point
void main(long long* pMemory, long long* pRegMMBLoc, long long* pFuncMem){
	

	HandleCode((struct Registers*) pRegMMBLoc, (uint8_t*) pMemory, (uint8_t*) pFuncMem);

}